#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import datetime
import json
import logging
import os
import re

# 3rd party imports
import numpy as np
import xarray as xr
from cdflib import cdfepoch
from dateutil import parser
from dateutil.rrule import DAILY, rrule
from scipy import integrate

from ..pyrf import read_cdf, ts_append

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def _list_files_tnr_l2(tint, data_path: str = "", tree: bool = False):
    """Find files in the L2 data repo corresponding to the target time
    interval.

    Parameters
    ----------
    tint : list
        Time interval
    data_path : str, Optional
        Path of MMS data. Default uses `pyrfu.solo.config.json`
    tree : bool, Optional
        Flag for tree structured data repos. Default is False.

    Returns
    -------
    files : list
        List of files corresponding to the parameters in the selected time
        interval

    """

    # Check path
    if not data_path:
        pkg_path = os.path.dirname(os.path.abspath(__file__))

        # Read the current version of the MMS configuration file
        with open(os.path.join(pkg_path, "config.json"), "r", encoding="utf-8") as fs:
            config = json.load(fs)

        data_path = os.path.normpath(config["local_data_dir"])
    else:
        data_path = os.path.normpath(data_path)

    # Make sure that the data path exists
    assert os.path.exists(data_path), f"{data_path} doesn't exist!!"

    files_out = []

    # directory and file name search patterns:
    # - assume directories are of the form: [data_path]/L2/thr/year/month/
    # - assume file names are of the form:
    #   solo_L2_rpw-tnr-surv-cdag_YYYYMMDD_version.cdf

    file_name = r"solo_L2_rpw-tnr-surv.*_([0-9]{8})_V[0-9]{2}.cdf"

    # Check tint
    assert isinstance(tint, (list, np.ndarray)), "tint must be array_like"
    assert len(tint) == 2, "tint must contain two elements"
    assert isinstance(tint[0], str), "tint[0] must be a string"
    assert isinstance(tint[1], str), "tint[1] must be a string"

    d_start = parser.parse(parser.parse(tint[0]).strftime("%Y-%m-%d"))
    until_ = parser.parse(tint[1]) - datetime.timedelta(seconds=1)
    days = rrule(DAILY, dtstart=d_start, until=until_)

    for date in days:
        if tree:
            local_dir = os.sep.join(
                [
                    data_path,
                    "L2",
                    "thr",
                    date.strftime("%Y"),
                    date.strftime("%m"),
                ],
            )
        else:
            local_dir = data_path

        if os.name == "nt":
            full_path = os.sep.join(
                [re.escape(local_dir) + os.sep, file_name],
            )
        else:
            full_path = os.sep.join([re.escape(local_dir), file_name])

        regex = re.compile(full_path)

        for root, _, files in os.walk(local_dir):
            for file in files:
                this_file = os.sep.join([root, file])

                matches = regex.match(this_file)
                if matches:
                    this_time = parser.parse(matches.groups()[0])
                    if d_start <= this_time <= until_:
                        if this_file not in files_out:
                            files_out.append(os.sep.join([local_dir, file]))

    # sort in time
    if len(files_out) > 1:
        files_out = sorted(files_out)

    return files_out


def read_tnr(tint, sensor: int = 4, data_path: str = "", tree: bool = False):
    r"""Read L2 data from TNR

    Parameters
    ----------
    tint : list
        Time interval
    sensor : int, Optional
        TNR sensor to be read:
            * 1: V1
            * 2: V2
            * 3: V3
            * 4: V1 - V2 (default)
            * 5: V2 - V3
            * 6: V3 - V1
            * 7: B
    data_path : str, Optional
        Path of MMS data. Default uses `pyrfu.solo.config.json`
    tree : bool, Optional
        Flag for tree structured data repos. Default is False.

    Returns
    -------
    out : xarray.DataArray
        Spectrum of the measured signals.

    Notes
    -----
    The script check if there are data from the two channel and put them
    together.

    """

    # Check input types
    assert isinstance(sensor, int), "sensor must integer"

    files = _list_files_tnr_l2(tint, data_path, tree)

    # Initialize spectrum output to None
    out = None

    for file in files:
        # Notify user
        logging.info("Loading %s...", os.path.split(file)[-1])

        data_l2 = read_cdf(file, tint)

        n_freqs = 4 * data_l2["tnr_band_freq"].shape[1]
        freq_tnr = np.reshape(data_l2["tnr_band_freq"].data, n_freqs) / 1000

        epoch_ = data_l2["epoch"].data
        auto1_ = data_l2["auto1"].data
        auto2_ = data_l2["auto2"].data
        sweep_ = data_l2["sweep_num"].data
        bande_ = data_l2["tnr_band"].data
        confg_ = data_l2["sensor_config"].data

        if sensor == 7:
            auto1_ = data_l2["magnetic_spectral_power1"].data
            auto2_ = data_l2["magnetic_spectral_power2"].data

        puntical = np.where(data_l2["front_end"].data == 1)[0]

        epoch_ = epoch_[puntical]
        confg_ = confg_[puntical, :]
        auto1_ = auto1_[puntical, :]
        auto2_ = auto2_[puntical, :]
        sweep_ = sweep_[puntical]
        bande_ = bande_[puntical]

        sweep_num = sweep_

        delta_sw = np.abs(sweep_[1:] - sweep_[:-1])

        xdelta_sw = np.where(delta_sw > 100)[0]
        if xdelta_sw.size:
            xdelta_sw = np.hstack([xdelta_sw, len(sweep_) - 1])
            nxdelta_sw = len(xdelta_sw)
            for inswn in range(nxdelta_sw - 1):
                idx_l, idx_r = [xdelta_sw[inswn] + 1, xdelta_sw[inswn + 1]]
                sweep_num[idx_l:idx_r] += sweep_num[xdelta_sw[inswn]]

        timet_ = cdfepoch.to_datetime(epoch_)

        sens0_, sens1_ = [np.where(confg_[:, i] == sensor)[0] for i in range(2)]

        if sens0_.size and sens1_.size:
            auto_calib = np.vstack([auto1_[sens0_, :], auto2_[sens1_, :]])
            sens_ = np.hstack([sens0_, sens1_])
            timet_ici = np.hstack([timet_[sens0_], timet_[sens1_]])
        elif sens0_.size:
            auto_calib = auto1_[sens0_, :]
            sens_ = sens0_
            timet_ici = timet_[sens0_]
        elif sens1_.size:
            auto_calib = auto2_[sens1_, :]
            sens_ = sens1_
            timet_ici = timet_[sens1_]

        else:
            raise ValueError("no data at all ?!?")

        ord_time = np.argsort(timet_ici)
        time_rr = timet_ici[ord_time]
        sens_ = sens_[ord_time]
        auto_calib = auto_calib[ord_time, :]

        bande_e = bande_[sens_]
        max_sweep = np.max(sweep_num[sens_])
        min_sweep = np.min(sweep_num[sens_])
        sweep_num = sweep_num[sens_]

        v_, time, sweepn_tnr = [[], [], []]

        for ind_sweep in range(min_sweep, max_sweep):
            v1_ = np.zeros(128)
            p_punt = np.where(sweep_num == ind_sweep)[0]
            if p_punt.size:
                for indband in range(p_punt.size):
                    idx_l = 32 * bande_e[p_punt[indband]]
                    idx_r = 32 * (bande_e[p_punt[indband]] + 1)
                    v1_[idx_l:idx_r] = auto_calib[p_punt[indband], :]

                if np.sum(v1_) > 0.0:
                    punt0_ = np.where(v1_ == 0.0)[0]
                    if punt0_.size:
                        v1_[punt0_] = np.nan
                    v_.append(v1_)
                    sweepn_tnr.append(sweep_num[p_punt[0]])

            if p_punt.size:
                time.append(time_rr[np.min(p_punt)])

        time = time[1:]
        v_ = np.stack(v_)

        # select frequencies lower than 100 kHz
        freq_tnr = freq_tnr[freq_tnr < 1e2]
        vp = v_[1:, : len(freq_tnr)]

        # Integration
        itg = integrate.trapz(vp, axis=1) / vp.shape[0]
        vp = vp - itg[:, None]

        out = ts_append(
            out,
            xr.DataArray(
                vp,
                coords=[np.stack(time), freq_tnr],
                dims=["time", "frequency"],
            ),
        )

        out = out[np.argsort(out.time.data)]

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import itertools

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from ..pyrf.normalize import normalize
from ..pyrf.resample import resample
from ..pyrf.ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _calc_angle(look_, vec):
    vec_hat = normalize(vec)
    theta_ = np.rad2deg(
        np.pi - np.arccos(np.sum(vec_hat.data * look_.data, axis=1)),
    )
    return theta_


def eis_pad(
    inp_allt,
    vec: xr.DataArray = None,
    energy: list = None,
    pa_width: int = 15,
):
    r"""Calculates Pitch Angle Distributions (PADs) using data from the MMS
    Energetic Ion Spectrometer (EIS)

    Parameters
    ----------
    inp_allt : xarray.Dataset
        Energy spectrum for all telescopes.
    vec : xarray.DataArray, Optional
        Axis with to compute pitch-angle. Default is X-GSE.
    energy : array_like, Optional
        Energy range to include in the calculation. Default is [55, 800].
    pa_width : int, Optional
        Size of the pitch angle bins, in degrees. Default is 15.

    Returns
    -------
    pa_flux : xarray.DataArray
        Pitch-angle angle distribution for every energy channels in the
        `energy` range.

    See Also
    --------
    pyrfu.mms.get_eis_allt

    """

    time_ = inp_allt.time.data

    if vec is None:
        vec = ts_vec_xyz(time_, np.tile(np.eye(1, 3), (len(time_), 1)))
    elif isinstance(vec, list):
        assert len(vec) == 3
        vec = ts_vec_xyz(time_, np.tile(vec, (len(time_), 1)))
    elif isinstance(vec, xr.DataArray):
        vec = resample(vec, inp_allt.time)

    if energy is None:
        energy = inp_allt.energy.data[[0, -1]]

    # set up the number of pa bins to create
    n_pabins = int(180.0 / pa_width)
    pa_label = list(180.0 * np.arange(0, n_pabins) / n_pabins + pa_width / 2.0)

    # Account for angular response (finite field of view) of instruments
    pa_hangw = 10.0  # deg
    delta_pa = pa_width / 2.0

    scopes = list(filter(lambda x: x.startswith("t"), inp_allt))

    pa_file = np.zeros([len(time_), len(scopes)])

    e_minu = inp_allt.energy.data + inp_allt.delta_energy_minus
    e_plus = inp_allt.energy.data + inp_allt.delta_energy_plus

    cond_low = np.logical_and(e_minu >= energy[0], e_minu <= energy[1])
    cond_hig = np.logical_and(e_plus >= energy[0], e_plus <= energy[1])
    ener_idx = np.where(np.logical_or(cond_low, cond_hig))[0]

    msg = "Energy range selected is not covered by the detector"
    assert np.sum(cond_low) != 0 and np.sum(cond_hig) != 0, msg

    flux_file = np.zeros([len(time_), len(scopes), len(ener_idx)])
    shape_ = [len(time_), n_pabins, len(ener_idx)]
    pa_flux = np.zeros(shape_)

    pa_flux[...] = np.nan

    for i_s, scope in enumerate(scopes):
        # get pa from each detector
        pa_file[:, i_s] = _calc_angle(inp_allt[f"look_{scope}"], vec)

        # get energy range of interest
        flux_file[:, i_s, :] = inp_allt[scope][:, ener_idx]

    flux_file[flux_file == 0] = np.nan

    for (i), (j, pa_lbl) in itertools.product(
        range(len(time_)),
        enumerate(pa_label),
    ):
        cond_ = np.logical_and(
            pa_file[i, :] + pa_hangw >= pa_lbl - delta_pa,
            pa_file[i, :] - pa_hangw < pa_lbl + delta_pa,
        )
        ind = np.where(cond_)[0]
        if ind.size != 0:
            pa_flux[i, j, :] = np.nanmean(flux_file[i, ind, :], axis=0)

    pa_flux = xr.DataArray(
        pa_flux,
        coords=[time_, pa_label, inp_allt.energy.data[ener_idx]],
        dims=["time", "theta", "energy"],
    )

    return pa_flux



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from scipy.stats import kurtosis

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def increments(inp, scale: int = 10):
    r"""Returns the increments of a time series.

    .. math:: y = |x_i - x_{i+s}|

    where :math:`s` is the scale.

    Parameters
    ----------
    inp : xarray.DataArray
        Input time series.
    scale : int, Optional
        Scale at which to compute the increments. Default is 10.

    Returns
    -------
    kurt : ndarray
        kurtosis of the increments, one per product, using the Fisher's
        definition (0 value for a normal distribution).
    result : xarray.DataArray
        An xarray containing the time series increments, one per
        product in the original time series.

    """

    assert isinstance(inp, xr.DataArray), "inp must be a xarray.DataArray"
    assert inp.ndim < 4, "inp must ber a scalar, vector or tensor"

    if inp.ndim == 1:
        data = inp.data[:, np.newaxis]
    else:
        data = inp.data

    # Compute the increments
    delta_inp = data[scale:, ...] - data[:-scale, ...]

    # Compute kurtosis of the increments
    kurt = kurtosis(delta_inp, axis=0, fisher=False)

    times, *comp = [inp.coords[dim].data for dim in inp.dims]

    result = xr.DataArray(
        np.squeeze(delta_inp),
        coords=[times[0 : len(delta_inp)], *comp],
        dims=inp.dims,
        attrs=inp.attrs,
    )

    return kurt, result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in import
import re
import warnings

# 3rd party imports
import numpy as np
from cdflib import CDF, cdfepoch

from ..pyrf.cdfepoch2datetime64 import cdfepoch2datetime64

# Local imports
from ..pyrf.datetime642iso8601 import datetime642iso8601
from ..pyrf.extend_tint import extend_tint
from ..pyrf.iso86012datetime64 import iso86012datetime64
from ..pyrf.time_clip import time_clip
from ..pyrf.ts_skymap import ts_skymap

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

# Keys of the global attributes to keep from CDF informations
Globkeys = [
    "CDF",
    "Version",
    "Encoding",
    "Checksum",
    "Compressed",
    "LeapSecondUpdated",
]


def _shift_epochs(file, epoch):
    r"""Shift times for particles."""

    epoch_shifted = epoch["data"].copy()

    try:
        delta_minus_var = {
            "data": file.varget(epoch["attrs"]["DELTA_MINUS_VAR"]),
            "attrs": file.varget(epoch["attrs"]["DELTA_MINUS_VAR"]),
        }
        delta_plus_var = {
            "data": file.varget(epoch["attrs"]["DELTA_PLUS_VAR"]),
            "attrs": file.varget(epoch["attrs"]["DELTA_PLUS_VAR"]),
        }

        delta_vars = [delta_minus_var, delta_plus_var]
        flags_vars = [1e3, 1e3]  # Time scaling conversion flags

        for i, delta_var in enumerate(delta_vars):
            if isinstance(delta_var["attrs"], dict) and "UNITS" in delta_var["attrs"]:
                if delta_var["attrs"]["UNITS"].lower() == "s":
                    flags_vars[i] = 1e3
                elif delta_var["attrs"]["UNITS"].lower() == "ms":
                    flags_vars[i] = 1e0
                else:
                    message = " units are not clear, assume s"
                    warnings.warn(message)
            else:
                message = "Epoch_plus_var/Epoch_minus_var units are not clear, assume s"
                warnings.warn(message)

        flag_minus, flag_plus = flags_vars
        t_offset = (
            delta_plus_var["data"] * flag_plus - delta_minus_var["data"] * flag_minus
        )
        t_offset = np.timedelta64(int(np.round(t_offset, 1) * 1e6 / 2), "ns")
        t_diff = (
            delta_plus_var["data"] * flag_plus - delta_minus_var["data"] * flag_minus
        )
        t_diff = np.timedelta64(int(np.round(t_diff, 1) * 1e6 / 2), "ns")
        t_diff_data = np.median(np.diff(epoch["data"])) / 2

        if t_diff_data != np.mean(t_diff):
            t_offset = t_diff_data

        epoch_shifted += t_offset

        return {"data": epoch_shifted, "attrs": epoch["attrs"]}

    except KeyError:
        return {"data": epoch_shifted, "attrs": epoch["attrs"]}


def _get_epochs(file, cdf_name, tint):
    r"""Get epochs form cdf and shift if needed."""

    depend0_key = file.varattsget(cdf_name)["DEPEND_0"]

    out = {
        "data": file.varget(depend0_key, starttime=tint[0], endtime=tint[1]),
    }

    if file.varinq(depend0_key).Data_Type_Description == "CDF_TIME_TT2000":
        try:
            out["data"] = cdfepoch2datetime64(out["data"])

            # Get epoch attributes
            out["attrs"] = file.varattsget(depend0_key)

            # Shift times if particle data
            is_part = re.search("^mms[1-4]_d[ei]s_", cdf_name)
            is_part = is_part or re.search("^mms[1-4]_hpca_", cdf_name)

            if is_part:
                out = _shift_epochs(file, out)

        except TypeError:
            pass

    return out


def get_dist(file_path, cdf_name, tint: list = None):
    r"""Read field named cdf_name in file and convert to velocity distribution
    function.

    Parameters
    ----------
    file_path : str
        Path of the cdf file.
    cdf_name : str
        Name of the target variable in the cdf file.
    tint : list of str, Optional
        Time interval.

    Returns
    -------
    out : xarray.Dataset
        Time series of the velocity distribution function if the target
        specie in the selected time interval.

    """

    tmmode = cdf_name.split("_")[-1]

    if "_dis_" in cdf_name:
        specie = "ions"
    elif "_des_" in cdf_name:
        specie = "electrons"
    else:
        raise AttributeError(
            "Couldn't get the particle species from file name!!",
        )

    # Check time interval type
    # Check time interval
    if tint is None:
        tint = ["1995-10-06T18:50:00.000000000", "2200-10-06T18:50:00.000000000"]
    elif isinstance(tint, (np.ndarray, list)):
        if isinstance(tint[0], np.datetime64):
            tint = datetime642iso8601(np.array(tint))
        elif isinstance(tint[0], str):
            tint = iso86012datetime64(
                np.array(tint),
            )  # to make sure it is ISO8601 ok!!
            tint = datetime642iso8601(np.array(tint))
        else:
            raise TypeError("Values must be in datetime64, or str!!")
    else:
        raise TypeError("tint must be array_like!!")

    # Extend time interval by 1s and convert time interval to epochs
    tint_org = tint.copy()
    tint = extend_tint(tint, [-1.0, 1.0])
    tint = list(datetime642iso8601(iso86012datetime64(np.array(tint))))
    tint = np.stack(list(map(cdfepoch.parse, tint)))

    with CDF(file_path) as file:
        # Get the relevant CDF file information (zVariables)
        cdf_infos = file.cdf_info()
        z_vars = cdf_infos.zVariables

        # Get the global attributes
        glob_attrs = file.globalattsget()
        glob_attrs = {**glob_attrs, **{"tmmode": tmmode, "species": specie}}

        # Get VDF zVariable attributes
        dist_attrs = file.varattsget(cdf_name)

        # Get CDF keys to Epoch, energy, azimuthal and elevation angle
        # zVariables
        dpnd_keys = [dist_attrs[f"DEPEND_{i:d}"] for i in range(4)]
        _, depend1_key, depend2_key, depend3_key = dpnd_keys

        # Get coordinates attributes
        coords_names = ["time", "phi", "theta", "energy"]
        coords_attrs = {n: file.varattsget(k) for n, k in zip(coords_names, dpnd_keys)}

        times = _get_epochs(file, cdf_name, tint)

        # If something time is None means that there is nothing interesting
        # in this file so leave!!
        if times["data"] is not None:
            times = times["data"]
        else:
            return None

        dist = file.varget(cdf_name, starttime=tint[0], endtime=tint[1])
        dist = np.transpose(dist, [0, 3, 1, 2])
        phi = file.varget(depend1_key, starttime=tint[0], endtime=tint[1])
        theta = file.varget(depend2_key)
        energy = file.varget(depend3_key, starttime=tint[0], endtime=tint[1])

        if tmmode == "brst":
            en0_name = "_".join(
                [
                    cdf_name.split("_")[0],
                    cdf_name.split("_")[1],
                    "energy0",
                    cdf_name.split("_")[-1],
                ],
            )
            en1_name = "_".join(
                [
                    cdf_name.split("_")[0],
                    cdf_name.split("_")[1],
                    "energy1",
                    cdf_name.split("_")[-1],
                ],
            )

            e_step_table_name = "_".join(
                [
                    cdf_name.split("_")[0],
                    cdf_name.split("_")[1],
                    "steptable_parity",
                    cdf_name.split("_")[-1],
                ],
            )

            step_table = file.varget(
                e_step_table_name,
                starttime=tint[0],
                endtime=tint[1],
            )

            if en0_name not in z_vars:
                if energy.ndim == 1:
                    energy0 = energy
                    energy1 = energy
                elif energy.shape[0] == 1:
                    energy0 = energy[0, :]
                    energy1 = energy[0, :]
                else:
                    energy0 = energy[1, :]
                    energy1 = energy[0, :]
            else:
                energy0 = file.varget(en0_name)
                energy1 = file.varget(en1_name)

            # Overwrite energy to make sure that energy0 and energy1
            # are used instead
            energy = np.tile(energy0, (len(step_table), 1))
            energy[step_table == 1] = np.tile(energy1, (int(np.sum(step_table)), 1))

        elif tmmode == "fast":
            phi = np.tile(phi, (len(times), 1))

            if energy.ndim == 1:
                energy0 = energy
                energy1 = energy
            elif energy.shape[0] == 1:
                energy0 = energy[0, :]
                energy1 = energy[0, :]
            else:
                energy0 = energy[1, :]
                energy1 = energy[0, :]

            step_table = np.zeros(len(times))

        else:
            raise ValueError("Invalid sampling mode!!")

        d_en_name = "_".join(
            [
                cdf_name.split("_")[0],
                cdf_name.split("_")[1],
                "energy_delta",
                cdf_name.split("_")[-1],
            ],
        )

        if d_en_name in z_vars:
            glob_attrs["delta_energy_plus"] = file.varget(
                d_en_name,
                starttime=tint[0],
                endtime=tint[1],
            )
            glob_attrs["delta_energy_minus"] = file.varget(
                d_en_name,
                starttime=tint[0],
                endtime=tint[1],
            )
        else:
            glob_attrs["delta_energy_plus"] = None
            glob_attrs["delta_energy_minus"] = None

        out = ts_skymap(
            times,
            dist,
            energy,
            phi,
            theta,
            energy0=energy0,
            energy1=energy1,
            esteptable=step_table,
            attrs=dist_attrs,
            coords_attrs=coords_attrs,
            glob_attrs=glob_attrs,
        )

    out = time_clip(out, tint_org)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import itertools

# Built-in imports
import warnings

# 3rd party imports
import numpy as np
import xarray as xr

from .feeps_active_eyes import feeps_active_eyes

# Local imports
from .feeps_pitch_angles import feeps_pitch_angles

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

# Angular response (finite field of view) of instruments electrons can use
# +/- 21.4 deg on each pitch angle as average response angle; ions can start
# with +/-10 deg, but both need to be further refined
angular_repsonse = {"electron": 21.4, "ion": 10}


def _pa_data_map(idx_maps, d_type, d_rate):
    pa_data_map = {}

    if d_rate == "srvy":
        pa_data_map[f"top-{d_type}"] = idx_maps[f"{d_type}-top"]
        pa_data_map[f"bottom-{d_type}"] = idx_maps[f"{d_type}-bottom"]
    else:
        # note: the following are indices of the top/bottom sensors in pa_data
        # they should be consistent with pa_dlimits.labels
        pa_data_map["top-electron"] = np.arange(9)
        pa_data_map["bottom-electron"] = np.arange(9, 18)

        # and ions:
        pa_data_map["top-ion"] = [0, 1, 2]
        pa_data_map["bottom-ion"] = [3, 4, 5]

    return pa_data_map


def _dpa_dflux(inp_dataset, pitch_angles, pa_data_map, energy, d_type, mms_id):
    pa_times = pitch_angles.time
    pa_data = pitch_angles.data

    trange = np.datetime_as_string(
        np.hstack([np.min(pa_times.data), np.max(pa_times.data)]),
        "ns",
    )

    eyes = feeps_active_eyes(inp_dataset.attrs, list(trange), mms_id)

    sensor_types = ["top", "bottom"]

    n_times = len(pa_times)
    n_top = len(pa_data_map[f"top-{d_type}"])
    n_bottom = len(pa_data_map[f"bottom-{d_type}"])

    dflux, dpa = [np.zeros([n_times, n_top + n_bottom]) for _ in range(2)]

    for s_type in sensor_types:
        pa_map = pa_data_map[f"{s_type}-{d_type}"]

        particle_idxs = [eye - 1 for eye in eyes[s_type]]

        for isen, sensor_num in enumerate(particle_idxs):
            var_name = f"{s_type}-{sensor_num + 1:d}"

            data = inp_dataset[var_name].data
            energies = inp_dataset[inp_dataset[var_name].dims[1]].data

            # remove any 0s before averaging
            data[data == 0] = "nan"

            # assumes all energies are NaNs if the first is
            if np.isnan(energies[0]):
                continue

            # energy indices to use:
            idx = np.where(
                np.logical_and(energies >= energy[0], energies <= energy[1]),
            )[0]

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                dflux[:, pa_map[isen]] = np.nanmean(data[:, idx], axis=1)

            dpa[:, pa_map[isen]] = pa_data[:, pa_map[isen]]

    # we need to replace the 0.0s left in after populating dpa with NaNs;
    # these 0.0s are left in there because these points aren't covered by
    # sensors loaded for this datatype/d_ratee
    dpa[dpa == 0] = "nan"

    return dpa, dflux


def _pa_flux(pa_times, pa_bins, pa_labels, dpa, dflux, d_type):
    n_pabins = len(pa_bins) - 1
    # Account for angular response
    dangresp = angular_repsonse[d_type]

    pa_flux = np.zeros([len(pa_times), int(n_pabins)])
    delta_pa = (pa_bins[1] - pa_bins[0]) / 2.0

    # Now loop through PA bins and time, find the telescopes where there is
    # data in those bins and average it up!
    for pa_idx, ipa in itertools.product(
        range(len(pa_times)),
        range(n_pabins),
    ):
        if not np.isnan(dpa[pa_idx, :][0]):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                ind = np.where(
                    (dpa[pa_idx, :] + dangresp >= pa_labels[ipa] - delta_pa)
                    & (dpa[pa_idx, :] - dangresp < pa_labels[ipa] + delta_pa),
                )

                if ind[0].size != 0:
                    if len(ind[0]) > 1:
                        pa_flux[pa_idx, ipa] = np.nanmean(
                            dflux[pa_idx, ind[0]],
                            axis=0,
                        )
                    else:
                        pa_flux[pa_idx, ipa] = dflux[pa_idx, ind[0]]

    pa_flux[pa_flux == 0] = "nan"  # fill any missed bins with NAN

    return pa_flux


def feeps_pad(
    inp_dataset,
    b_bcs,
    bin_size: float = 16.3636,
    energy: list = None,
):
    r"""Compute pitch angle distribution using FEEPS data.

    Parameters
    ----------
    inp_dataset : xarray.Dataset
        Energy spectrum of all eyes.
    b_bcs : xarray.DataArray
        Time series of the magnetic field in spacecraft coordinates.
    bin_size : float, Optional
        Width of the pitch angles bins. Default is 16.3636.
    energy : array_like, Optional
        Energy range of particles. Default is [70., 600.]

    Returns
    -------
    pad : xarray.DataArray
        Time series of the pitch angle distribution.

    """

    if energy is None:
        energy = [70.0, 600.0]

    assert energy[0] > 32.0, "Please use a starting energy of 32 keV or above"

    time = inp_dataset.time.data
    attrs = inp_dataset.attrs
    mms_id, d_type, d_rate = list(map(attrs.get, ["mmsId", "dtype", "tmmode"]))

    assert d_rate in ["srvy", "brst"]
    assert d_type in ["electron", "ion"]

    n_pabins = int(180 / bin_size)
    pa_bins = [180.0 * pa_bin / n_pabins for pa_bin in range(n_pabins + 1)]
    pa_labels = [pa_bin + bin_size / 2.0 for pa_bin in pa_bins[:-1]]

    pitch_angles, idx_maps = feeps_pitch_angles(inp_dataset, b_bcs)

    pa_data_map = _pa_data_map(idx_maps, d_type, d_rate)

    dpa, dflux = _dpa_dflux(
        inp_dataset,
        pitch_angles,
        pa_data_map,
        energy,
        d_type,
        mms_id,
    )

    pa_flux = _pa_flux(
        pitch_angles.time,
        pa_bins,
        pa_labels,
        dpa,
        dflux,
        d_type,
    )

    pad = xr.DataArray(
        pa_flux,
        coords=[time, pa_labels],
        dims=["time", "theta"],
        attrs=attrs,
    )

    return pad



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import matplotlib.dates as dates
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.backend_bases import MouseButton
from matplotlib.widgets import Button

from ..plot import plot_line

# Local imports
from .avg_4sc import avg_4sc
from .iso86012unix import iso86012unix
from .resample import resample
from .t_eval import t_eval

__author__ = "Atlas Silverhult"
__email__ = "atlas.silverhult.9977@student.uu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

__all__ = ["timing_gui"]


def _get_disc_vel(t, tau, r_list):
    # tetrahedron center
    rstar = t_eval(avg_4sc(r_list), t)
    # position vectors with origin in tetrahedron center
    r_list_star = [t_eval(r_, t) - rstar for r_ in r_list]

    # get position tensor Rstar (1)
    r_start = sum([np.outer(r_, r_.T) for r_ in r_list_star])

    # generelazied reciprocal vectors q (3) (q=k)
    q_list = [np.linalg.inv(r_start) @ rstar_.T.data for rstar_ in r_list_star]

    # get slowness vector (17)
    m = sum([q_list[i] * tau[i] for i in range(4)])
    # tidy
    m = m.T[0]

    # normal vector from (9)
    n = m / np.linalg.norm(m)

    # disc vel from (9)
    v = 1 / np.linalg.norm(m)

    return m, n, v


def _print_output(tau, m, n, v, header):
    pad = 15
    topstr = "-" * pad + f"{header}" + "-" * pad
    print("\n" + topstr)
    print(f"time diff from {header} (seconds):\n{tau}")
    print("\nn: ", n)
    print("m [s/km]: ", m)
    print("V [km/s]: ", v)
    print("-" * len(topstr))


def timing_gui(b_list: list, r_list: list, comp: int = 0):
    r"""GUI to perform manual timing analysis by clicking. Limited to 4 spacecraft only.
    The return of this function is a callback to the GUI object and class attributes
    like the slowness vector are accesable through this callback by the methods
    'get_slowness', 'get_normal' and 'get_vel'.

    Usage:
    Before cicking the 'Start timing' button, right clicking will place a vertical
    line to mark a timestamp of interest. After clicking 'Start timing' the four
    signals will be highlighted one at a time, indicating which signal should be
    clicked. For each signal, use RIGHT CLICK to mark the time and move on to the next
    one. By using LEFT CLICK, one can undo the selection for the previously selected
    signal(s). Once the times have been marked for each signal, the 'Calculate normal'
    button will calculate the speed and normal vector based on the input times.

    Parameters
    ----------
    b_list : list of xarray.DataArray
        List of magnetic field time series for each sc.
    r_list : list of xarray.DataArray
        List of coordinate time series for each sc.
    comp : int, Optional
        Component to use. Default uses first component.

    Returns
    -------
    timing_gui_callback : TYPE
        Returns TimingGui object to access attributes. In order to keep
        GUI responsive and interactive, a reference to this object is needed.


    """

    # Resample coordinate time series to magnetic field time line
    r_list = [resample(r_, b_) for r_, b_ in zip(r_list, b_list)]
    b_list_x = [b_[:, comp] for b_ in b_list]

    timing_gui_callback = TimingGui(b_list_x, r_list)

    return timing_gui_callback


class TimingGui:
    def __init__(self, b, r):
        # List of time series data
        self.b = b
        self.r = r

        # Output quantities - to be calculated
        self.n = None
        self.m = None
        self.V = None

        # Figures and GUI
        self.fig = self.init_fig(self.b)
        self.index = 0
        self.timing_started = False

        self.calc_button = Button(
            self.fig.axes[4], "Calculate normal", color="0.85", hovercolor="0.65"
        )
        self.calc_button.on_clicked(self.calc_vel_button)

        self.timing_button = Button(
            self.fig.axes[3], "Begin timing", color="0.85", hovercolor="0.65"
        )
        self.timing_button.on_clicked(self.begin_timing)

        plt.connect("button_press_event", self.onclick)

    @staticmethod
    def init_fig(b_list):
        b1, b2, b3, b4 = b_list

        fig = plt.figure(figsize=(16, 9))
        legend_options = dict(
            ncol=1,
            loc="center left",
            frameon=True,
            framealpha=1,
            bbox_to_anchor=(1, 0.5),
        )

        gs = gridspec.GridSpec(
            3, 5, bottom=0.05, top=0.9, left=0.05, right=0.9, hspace=0.4, wspace=0.35
        )
        # Time series data
        ax1 = fig.add_subplot(gs[0, :])

        # MVA transformed frame
        fig.add_subplot(gs[1, :])

        # Min/Max hodogram
        ax3 = fig.add_subplot(gs[2, :3])
        fig.add_subplot(gs[2, 3:4])
        fig.add_subplot(gs[2, 4:])

        ax1.xaxis.set_major_locator(dates.AutoDateLocator(minticks=30, maxticks=50))
        ax3.axis("off")

        # Plot
        plot_line(ax1, b1, color="black")
        plot_line(ax1, b2)
        plot_line(ax1, b3)
        plot_line(ax1, b4)
        b_labels = ["sc1", "sc2", "sc3", "sc4"]
        ax1.legend(b_labels, **legend_options)

        return fig

    def display_text(self, m, n, v):
        ax3 = self.fig.axes[2]
        for txt in ax3.texts:
            txt.set_visible(False)
        fontsize = 20
        ax3.text(0, 1, f"V [km/s]: {np.round(v ,3)}", fontsize=fontsize, va="top")
        ax3.text(0, 0.5, f"m [s/km]: {np.round(m ,3)}", fontsize=fontsize, va="center")
        ax3.text(0, 0, f"n: {np.round(n ,3)}", fontsize=fontsize, va="bottom")

        self.fig.canvas.draw_idle()

    def begin_timing(self, event):
        # Called if button pressed
        for line in self.fig.axes[1].get_lines():
            line.remove()

        self.timing_started = True
        self.time_list = np.zeros(4)
        self.index = 0
        self.update_selection()
        self.fig.canvas.draw_idle()

    def onclick(self, event):
        # Add time click to list if
        if self.timing_started and event.inaxes == self.fig.axes[0]:
            self.store_t(event)

        # Undo last time click if right mouse button clicked
        elif (
            not self.timing_started
            and event.inaxes == self.fig.axes[0]
            and event.button is MouseButton.RIGHT
        ):
            for line in self.fig.axes[0].lines[4:]:
                line.remove()
            self.fig.axes[0].axvline(
                event.xdata, -1, 1, linestyle="--", linewidth=2, color="black"
            )
            self.fig.canvas.draw_idle()

    def store_t(self, event):
        if event.button is MouseButton.LEFT and event.inaxes == self.fig.axes[0]:
            t_click = event.xdata
            # print("Matplot time: ",t_click)

            if self.index < 4:
                self.time_list[self.index] = t_click
                self.index += 1

            elif self.index >= 4:
                print("Time list full!")

        if event.button is MouseButton.RIGHT and event.inaxes == self.fig.axes[0]:
            if self.index < 1:
                print("List already empty")
            else:
                self.index -= 1
                self.time_list[self.index] = 0

        self.update_selection()

    def update_selection(self):
        """
        Highlight one line at a time untill all four have been clicked and
        the shifted signals are plotted by calling self.plot_shifted.

        """
        current = self.index
        lines = self.fig.axes[0].lines
        if current == 4:
            for line in lines[:4]:
                line.set_alpha(1)
            self.plot_shifted(self.time_list)
            self.timing_started = False

        else:
            for line in lines[:4]:
                line.set_alpha(0.2)
            lines[current].set_alpha(1)

        self.fig.canvas.draw_idle()

    def plot_shifted(self, time_list):
        # Shift time series relative sc1 (t0) and plot it in the second panel

        t0 = np.datetime64(dates.num2date(time_list[0]))
        time_difs = [(t0 - np.datetime64(dates.num2date(t))) for t in time_list]
        # Plot shifted
        legend_options = dict(
            ncol=1,
            loc="center left",
            frameon=True,
            framealpha=1,
            bbox_to_anchor=(1, 0.5),
        )
        ax2 = self.fig.axes[1]
        colors = ["black", "blue", "green", "red"]
        for i, b_ in enumerate(self.b):
            b_shifted = b_.assign_coords(time=b_.time.data + time_difs[i])
            plot_line(ax2, b_shifted, color=colors[i], alpha=1)

        b_labels = ["sc1", "sc2", "sc3", "sc4"]

        ax2.legend(b_labels, **legend_options)

    def calc_vel_button(self, event):
        if self.index < 4:
            raise TypeError(f"Not enough times in time list, {self.index} out of 4")

        else:
            # Convert clicked locations to list of np.datetime64 values
            time_list64 = [(np.datetime64(dates.num2date(t))) for t in self.time_list]

            # Calculate time differences and covert to seconds
            tau = [(t - time_list64[0]).astype("float") * 1e-6 for t in time_list64]

            # Assign copy of list of position time series to new variable
            r_list = self.r.copy()

            # Time of first click
            t = iso86012unix([time_list64[0]])

            # Get disc velocity at time of first click
            m, n, v = _get_disc_vel(t, tau, r_list)

            # Set class variables to access from timing_gui object
            self.m, self.n, self.V = m, n, v

            # Display results in GUI window
            self.display_text(m, n, v)

            # Print to console
            _print_output(tau, m, n, v, "Clicking")

    def get_slowness(self):
        if self.m is None:
            raise Exception("Slowness vector has not been calculated yet")
        else:
            return self.m

    def get_normal(self):
        if self.n is None:
            raise Exception("Normal has not been calculated yet")
        else:
            return self.n

    def get_vel(self):
        if self.V is None:
            raise Exception("Velocity has not been calculated yet")
        else:
            return self.V



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy import constants

from ..pyrf.dec_par_perp import dec_par_perp
from ..pyrf.norm import norm
from ..pyrf.resample import resample
from ..pyrf.trace import trace
from ..pyrf.ts_scalar import ts_scalar

# Local imports
from .rotate_tensor import rotate_tensor

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def make_model_vdf(
    vdf,
    b_xyz,
    sc_pot,
    n_s,
    v_xyz,
    t_xyz,
    isotropic: bool = False,
):
    r"""Make a general bi-Maxwellian distribution function based on particle
    moment data in the same format as PDist.

    Parameters
    ----------
    vdf : xarray.Dataset
        Particle distribution (skymap).
    b_xyz : xarray.DataArray
        Time series of the background magnetic field.
    sc_pot : xarray.DataArray
        Time series of the spacecraft potential.
    n_s : xarray.DataArray
        Time series of the number density of specie s.
    v_xyz : xarray.DataArray
        Time series of the bulk velocity.
    t_xyz : xarray.DataArray
        Time series of the temperature tensor.
    isotropic : bool, Optional
        Flag to make an isotropic model distribution. Default is False.

    Returns
    -------
    model_vdf : xarray.Dataset
        Distribution function in the same format as vdf.

    See also
    --------
    pyrfu.mms.calculate_epsilon : Calculates epsilon parameter using model distribution.

    Examples
    --------
    >>> from pyrfu.mms import get_data, make_model_vdf

    Define time interval

    >>> tint_brst = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]

    Load magnetic field and spacecraft potential

    >>> b_dmpa = get_data("b_dmpa_fgm_brst_l2", tint_brst, 1)
    >>> scpot = get_data("V_edp_brst_l2", tint_brst, 1)

    Load electron velocity distribution function

    >>> vdf_e = get_data("pde_fpi_brst_l2", tint_brst, 1)

    Load moments of the electron velocity distribution function

    >>> n_e = get_data("ne_fpi_brst_l2", tint_brst, 1)
    >>> v_xyz_e = get_data("ve_dbcs_fpi_brst_l2", tint_brst, 1)
    >>> t_xyz_e = get_data("te_dbcs_fpi_brst_l2", tint_brst, 1)

    Compute model electron velocity distribution function

    >>> vdf_m_e = make_model_vdf(vdf_e, b_xyz, scpot, n_e, v_xyz_e, t_xyz_e)

    """

    assert vdf.attrs["species"][0].lower() in ["i", "e"], "Invalid specie"

    # Check that VDF and moments have the same timeline
    message = "VDF and moments have different times."
    assert np.abs(np.median(np.diff(vdf.time.data - n_s.time.data))) == 0, message

    # Resample b_xyz and sc_pot to particle data resolution
    b_xyz, sc_pot = [resample(b_xyz, n_s), resample(sc_pot, n_s)]

    # Define directions based on b_xyz and v_xyz, calculate relevant
    # temperatures. N.B makes final distribution gyrotropic
    t_xyzfac = rotate_tensor(t_xyz, "fac", b_xyz, "pp")

    if isotropic:
        t_para = trace(t_xyzfac) / 3
        t_ratio = ts_scalar(
            t_xyzfac.time.data,
            np.ones(len(t_xyzfac.time.data)),
        )
    else:
        t_para = t_xyzfac[:, 0, 0]
        t_ratio = t_xyzfac[:, 0, 0] / t_xyzfac[:, 1, 1]

    v_para, v_perp, _ = dec_par_perp(v_xyz, b_xyz)

    v_perp_mag, b_xyz_mag = [norm(v_perp), norm(b_xyz)]
    v_perp_dir, b_xyz_dir = [v_perp / v_perp_mag, b_xyz / b_xyz_mag]

    # Define constants
    q_e = constants.elementary_charge

    # Check whether particles are electrons or ions
    if vdf.attrs["species"][0].lower() == "e":
        p_mass = constants.electron_mass
    else:
        p_mass = constants.proton_mass
        sc_pot.data = -1.0 * sc_pot.data

    # Convert moments to SI units
    vth_para = np.sqrt(2 * t_para.data * q_e / p_mass)

    v_perp_mag_data = 1e3 * v_perp_mag.data
    v_para_data = 1e3 * v_para.data
    n_s_data = 1e6 * n_s.data

    # Defines dimensions of array below
    n_ti = len(vdf.time)
    n_en = len(vdf.energy.data[0, :])
    n_ph, n_th = [len(angle) for angle in [vdf.phi[0, :], vdf.theta]]

    # Get energy array
    energy = vdf.energy

    # Define Cartesian coordinates
    x_mat, y_mat, z_mat = [np.zeros((n_ti, n_ph, n_th)) for _ in range(3)]

    r_mat = np.zeros((n_ti, n_en))

    for i in range(n_ti):
        x_mat[i, ...] = np.outer(
            -np.cos(np.deg2rad(vdf.phi.data[i, :])),
            np.sin(np.deg2rad(vdf.theta.data)),
        )
        y_mat[i, ...] = np.outer(
            -np.sin(np.deg2rad(vdf.phi.data[i, :])),
            np.sin(np.deg2rad(vdf.theta.data)),
        )
        z_mat[i, ...] = np.outer(
            -np.ones(n_ph),
            np.cos(np.deg2rad(vdf.theta.data)),
        )
        r_mat[i, ...] = np.real(
            np.sqrt(2 * (energy[i, :] - sc_pot.data[i]) * q_e / p_mass),
        )

    r_mat[r_mat == 0] = 0.0

    # Define rotation vectors based on B and Ve directions
    r_x = v_perp_dir.data
    r_y = np.cross(b_xyz_dir.data, v_perp_dir.data)
    r_z = b_xyz_dir.data

    # Rotated coordinate system for computing bi-Maxwellian distribution
    x_p, y_p, z_p = [np.zeros((n_ti, n_ph, n_th)) for _ in range(3)]

    for i in range(n_ti):
        x_p[i, ...] = (
            x_mat[i, ...] * r_x[i, 0]
            + y_mat[i, ...] * r_x[i, 1]
            + z_mat[i, ...] * r_x[i, 2]
        )

        y_p[i, ...] = (
            x_mat[i, ...] * r_y[i, 0]
            + y_mat[i, ...] * r_y[i, 1]
            + z_mat[i, ...] * r_y[i, 2]
        )

        z_p[i, ...] = (
            x_mat[i, ...] * r_z[i, 0]
            + y_mat[i, ...] * r_z[i, 1]
            + z_mat[i, ...] * r_z[i, 2]
        )

    # Make 4D position matrix
    x_p = np.transpose(np.tile(x_p, [n_en, 1, 1, 1]), [1, 0, 2, 3])
    y_p = np.transpose(np.tile(y_p, [n_en, 1, 1, 1]), [1, 0, 2, 3])
    z_p = np.transpose(np.tile(z_p, [n_en, 1, 1, 1]), [1, 0, 2, 3])
    r_mat = np.transpose(np.tile(r_mat, [n_ph, n_th, 1, 1]), [2, 3, 0, 1])

    # Construct bi-Maxwellian distribution function
    bi_max_dist = np.zeros(r_mat.shape)

    for i in range(n_ti):
        coeff = (
            n_s_data[i]
            * t_ratio.data[i]
            / (np.sqrt(np.pi**3) * vth_para.data[i] ** 3)
        )

        bi_max_temp = coeff * np.exp(
            -((x_p[i, ...] * r_mat[i, ...] - v_perp_mag_data[i]) ** 2)
            / (vth_para.data[i] ** 2)
            * t_ratio.data[i],
        )
        bi_max_temp = bi_max_temp * np.exp(
            -((y_p[i, ...] * r_mat[i, ...]) ** 2)
            / (vth_para.data[i] ** 2)
            * t_ratio.data[i],
        )
        bi_max_temp = bi_max_temp * np.exp(
            -((z_p[i, ...] * r_mat[i, ...] - v_para_data[i]) ** 2)
            / (vth_para.data[i] ** 2),
        )

        bi_max_dist[i, ...] = bi_max_temp

    # Make modelPDist file for output
    model_vdf = vdf.copy()
    model_vdf.data.data = bi_max_dist
    model_vdf.data.data *= 1e18
    model_vdf.attrs["UNITS"] = "s^3/km^6"

    return model_vdf



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import warnings

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _check_spin(spin):
    _, spin_inds = np.unique(spin, return_index=True)

    len_spin = np.max(np.diff(spin_inds))

    # Check to see if we have a complete first spin,
    # if not then go to the second spin
    if np.where(spin == spin[spin_inds[0]])[0].size != len_spin:
        spin_inds = spin_inds[1:]

    # Check to see if the last one is complete,
    # if not then go to the second spin
    if np.where(spin == spin[spin_inds[-1]])[0].size != len_spin:
        spin_inds = spin_inds[:-1]

    return spin_inds, len_spin


def eis_spin_avg(eis_allt, method: str = "mean"):
    r"""Calculates spin-averaged fluxes for the EIS instrument.

    Parameters
    ----------
    eis_allt : xarray.Dataset
        Dataset of the fluxes of all 6 telescopes.
    method : {"mean", "sum"}
        Method.
    Returns
    -------
    spin_avg_fluxes : xarray.Dataset
        Spin-averaged fluxes for all 6 telescopes.

    See Also
    --------
    pyrfu.mms.get_eis_allt, pyrfu.mms.eis_omni

    Examples
    --------
    >>> from pyrfu import mms

    Define spacecraft index and time interval

    >>> tint = ["2017-07-23T16:10:00", "2017-07-23T18:10:00"]
    >>> ic = 2

    Get EIS ExTOF all 6 telescopes fluxes

    >>> extof_allt = mms.get_eis_allt("flux_extof_proton_srvy_l2", tint, ic)

    Spin average all 6 telescopes fluxes

    >>> extof_allt_despin = mms.eis_spin_avg(extof_allt)

    """

    assert method.lower() in ["mean", "sum"]

    spin_nums = eis_allt.spin

    spin_starts, len_spin = _check_spin(spin_nums.data)

    scopes = list(filter(lambda x: x[0] == "t", eis_allt))

    spin_avg_flux = {}

    for scope in scopes:
        scope_data = eis_allt[scope]

        time_recs = scope_data.time.data[spin_starts + len_spin // 2]
        energies_ = scope_data.energy.data
        flux_data = scope_data.data

        flux_avg = np.zeros([len(spin_starts), len(energies_)])

        for i, spin_strt in enumerate(spin_starts):
            t_inds = np.where(spin_nums.data == spin_nums.data[spin_strt])[0]

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)

                if method.lower() == "sum":
                    # Sum (for counts)
                    flux_avg[i, :] = np.nansum(flux_data[t_inds, :], axis=0)

                else:
                    # Average (for differential partile flux, differential
                    # energy flux, phase-space density, etc.)
                    flux_avg[i, :] = np.nanmean(flux_data[t_inds, :], axis=0)

        spin_avg_flux[scope] = xr.DataArray(
            flux_avg,
            coords=[time_recs, energies_],
            dims=["time", "energy"],
            attrs=scope_data.attrs,
        )

    # Get spin index
    spin_avg_flux["spin"] = eis_allt["spin"][spin_starts + len_spin // 2]

    # Construct Dataset
    out = xr.Dataset(spin_avg_flux, attrs=eis_allt.attrs)
    out.time.attrs = eis_allt.time.attrs
    out.energy.attrs = eis_allt.energy.attrs

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd oarty imports
import numpy as np

# Local imports
from ..pyrf.calc_fs import calc_fs
from ..pyrf.ts_scalar import ts_scalar
from ..pyrf.ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def fft_bandpass(inp, f_min, f_max):
    r"""Perform simple bandpass using FFT - returns fields between with
    ``f_min`` < f < ``f_max``.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series to be bandpassed filtered.
    f_min : float or int
        Minimum frequency of filter, f < ``f_min`` are removed.
    f_max : float or int
        Maximum frequency of filter, f > ``f_max`` are removed.

    Returns
    -------
    out : xarray.DataArray
        Time series of the bandpassed filtered data.

    Notes
    -----
    Can be some spurius effects near boundary. Can take longer interval then
    use tlim to remove.

    Examples
    --------
    >>> from pyrfu import mms

    Define time interval

    >>> tint = ["2017-07-23T16:54:24.000", "2017-07-23T17:00:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load Electric Field

    >>> e_xyz = mms.get_data("e_gse_edp_brst_l2", tint, mms_id)

    Bandpass filter

    >>> e_xyz_bp = mms.fft_bandpass(e_xyz, 1e1, 1e2)

    """

    inp_time, inp_data = [inp.time.data, inp.data]

    # Make sure number of elements is an even number, if odd remove last
    # element to make an even
    # number
    n_els = len(inp_data)

    if n_els % 2:
        inp_time = inp_time[:-1]
        inp_data = inp_data[:-1, :]

        n_els = len(inp_data)

    try:
        num_fields = inp_data.shape[1]
    except IndexError:
        num_fields = 1
        inp_data = inp_data[:, np.newaxis]

    # Set NaN values to zero so FFT works
    idx_nans = np.isnan(inp_data)
    inp_data[idx_nans] = 0.0

    # Bandpass filter field data
    f_sam = calc_fs(inp)
    f_nyq = f_sam / 2
    frequencies = np.linspace(-f_nyq, f_nyq, n_els)

    # FFT and remove frequencies
    for i in range(num_fields):
        inp_temp = np.fft.fft(inp_data[:, i])
        inp_temp = np.fft.fftshift(inp_temp)

        inp_temp[np.abs(frequencies) < f_min] = 0
        inp_temp[np.abs(frequencies) > f_max] = 0

        inp_temp = np.fft.ifftshift(inp_temp)
        inp_data[:, i] = np.fft.ifft(inp_temp)

    # Put back original NaNs
    inp_data[idx_nans] = np.nan

    # Return data in the same format as input
    if num_fields == 1:
        out = ts_scalar(inp_time, inp_data, attrs=inp.attrs)
    elif num_fields == 3:
        out = ts_vec_xyz(inp_time, inp_data, attrs=inp.attrs)
    else:
        raise ValueError("Invalid shape")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

# Local imports
from .find_closest import find_closest

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def corr_deriv(inp0, inp1, flag: bool = False):
    r"""Correlate the derivatives of two time series

    Parameters
    ----------
    inp0 : xarray.DataArray
        Time series of the first to variable to correlate with.
    inp1 : xarray.DataArray
        Time series of the second to variable to correlate with.
    flag : bool, Optional
        Flag if False (default) returns time instants of common highest first
        and second derivatives. If True returns time instants of common
        highest first derivative and zeros crossings.

    Returns
    -------
    t1_d, t2_d : ndarray
        Time instants of common highest first derivatives.
    t1_dd, t2_dd : ndarray
        Time instants of common highest second derivatives or zero crossings.

    """

    # 1st derivative
    tx1 = inp0.time.data.astype(np.int64) * 1e-9
    inp0 = inp0.data
    dtx1 = tx1[:-1] + 0.5 * np.diff(tx1)
    dx1 = np.diff(inp0)

    tx2 = inp1.time.data.astype(np.int64) * 1e-9
    inp1 = inp1.data
    dtx2 = tx2[:-1] + 0.5 * np.diff(tx2)
    dx2 = np.diff(inp1)

    ind_zeros1 = np.where(np.sign(dx1[:-1] * dx1[1:]) < 0)[0]
    if ind_zeros1 == 0:
        ind_zeros1 = ind_zeros1[1:]

    ind_zeros2 = np.where(np.sign(dx2[:-1] * dx2[1:]) < 0)[0]
    if ind_zeros2 == 0:
        ind_zeros2 = ind_zeros2[1:]

    ind_zeros1_p = np.where(dx1[ind_zeros1 - 1] - dx1[ind_zeros1] > 0)[0]
    ind_zeros2_p = np.where(dx2[ind_zeros2 - 1] - dx2[ind_zeros2] > 0)[0]

    ind_zeros1_m = np.where(dx1[ind_zeros1 - 1] - dx1[ind_zeros1] < 0)[0]
    ind_zeros2_m = np.where(dx2[ind_zeros2 - 1] - dx2[ind_zeros2] < 0)[0]

    ind1_p = ind_zeros1[ind_zeros1_p]
    ind1_m = ind_zeros1[ind_zeros1_m]

    t_zeros1_p = dtx1[ind1_p] + (dtx1[ind1_p + 1] - dtx1[ind1_p]) / (
        1 + np.abs(dx1[ind1_p + 1]) / np.abs(dx1[ind1_p])
    )
    t_zeros1_m = dtx1[ind1_m] + (dtx1[ind1_m + 1] - dtx1[ind1_m]) / (
        1 + np.abs(dx1[ind1_m + 1]) / np.abs(dx1[ind1_m])
    )

    ind2_p = ind_zeros2[ind_zeros2_p]
    ind2_m = ind_zeros2[ind_zeros2_m]

    t_zeros2_p = dtx2[ind2_p] + (dtx2[ind2_p + 1] - dtx2[ind2_p]) / (
        1 + np.abs(dx2[ind2_p + 1]) / np.abs(dx2[ind2_p])
    )
    t_zeros2_m = dtx2[ind2_m] + (dtx2[ind2_m + 1] - dtx2[ind2_m]) / (
        1 + np.abs(dx2[ind2_m + 1]) / np.abs(dx2[ind2_m])
    )

    # Remove repeating points
    t_zeros1_p = np.delete(t_zeros1_p, np.where(np.diff(t_zeros1_p) == 0)[0])
    t_zeros2_p = np.delete(t_zeros2_p, np.where(np.diff(t_zeros2_p) == 0)[0])

    # Define identical pairs of two time axis
    t1_d_p, t2_d_p, _, _ = find_closest(t_zeros1_p, t_zeros2_p)
    t1_d_m, t2_d_m, _, _ = find_closest(t_zeros1_m, t_zeros2_m)

    t1_d = np.vstack([t1_d_p, t1_d_m])
    t1_d = t1_d[t1_d[:, 0].argsort(), 0]

    t2_d = np.vstack([t2_d_p, t2_d_m])
    t2_d = t2_d[t2_d[:, 0].argsort(), 0]

    if flag:
        # zero crossings
        ind_zeros1 = np.where(np.sign(inp0[:-1] * inp0[1:]) < 0)[0]
        ind_zeros2 = np.where(np.sign(inp1[:-1] * inp1[1:]) < 0)[0]

        ind_zeros1 = np.delete(ind_zeros1, np.where(ind_zeros1 == 1)[0])
        ind_zeros2 = np.delete(ind_zeros2, np.where(ind_zeros2 == 1)[0])

        ind_zeros1_p = np.where(inp0[ind_zeros1 - 1] - inp0[ind_zeros1] > 0)[0]
        ind_zeros2_p = np.where(inp1[ind_zeros2 - 1] - inp1[ind_zeros2] > 0)[0]

        ind_zeros1_m = np.where(inp0[ind_zeros1 - 1] - inp0[ind_zeros1] < 0)[0]
        ind_zeros2_m = np.where(inp1[ind_zeros2 - 1] - inp1[ind_zeros2] < 0)[0]

        ind1_p = ind_zeros1[ind_zeros1_p]
        ind1_m = ind_zeros1[ind_zeros1_m]

        t_zeros1_p = tx1[ind1_p] + (tx1[ind1_p + 1] - tx1[ind1_p]) / (
            1 + np.abs(inp0[ind1_p + 1]) / np.abs(inp0[ind1_p])
        )
        t_zeros1_m = tx1[ind1_m] + (tx1[ind1_m + 1] - tx1[ind1_m]) / (
            1 + np.abs(inp0[ind1_m + 1]) / np.abs(inp0[ind1_m])
        )

        ind2_p = ind_zeros2[ind_zeros2_p]
        ind2_m = ind_zeros2[ind_zeros2_m]

        t_zeros2_p = tx2[ind2_p] + (tx2[ind2_p + 1] - tx2[ind2_p]) / (
            1 + np.abs(inp1[ind2_p + 1]) / np.abs(inp1[ind2_p])
        )
        t_zeros2_m = tx2[ind2_m] + (tx2[ind2_m + 1] - tx2[ind2_m]) / (
            1 + np.abs(inp1[ind2_m + 1]) / np.abs(inp1[ind2_m])
        )

    else:
        # 2nd derivative
        dd_tx1 = dtx1[:-1] + 0.5 * np.diff(dtx1)
        ddx1 = np.diff(dx1)

        dd_tx2 = dtx2[:-1] + 0.5 * np.diff(dtx2)
        ddx2 = np.diff(dx2)

        ind_zeros1 = np.where(np.sign(ddx1[:-1] * ddx1[1:]) < 0)[0]
        ind_zeros2 = np.where(np.sign(ddx2[:-1] * ddx2[1:]) < 0)[0]

        ind_zeros1 = np.delete(ind_zeros1, np.where(ind_zeros1 == 1)[0])
        ind_zeros2 = np.delete(ind_zeros2, np.where(ind_zeros2 == 1)[0])

        ind_zeros1_p = np.where(ddx1[ind_zeros1 - 1] - ddx1[ind_zeros1] > 0)[0]
        ind_zeros2_p = np.where(ddx2[ind_zeros2 - 1] - ddx2[ind_zeros2] > 0)[0]

        ind_zeros1_m = np.where(ddx1[ind_zeros1 - 1] - ddx1[ind_zeros1] < 0)[0]
        ind_zeros2_m = np.where(ddx2[ind_zeros2 - 1] - ddx2[ind_zeros2] < 0)[0]

        ind1_p = ind_zeros1[ind_zeros1_p]
        ind1_m = ind_zeros1[ind_zeros1_m]

        t_zeros1_p = dd_tx1[ind1_p] + (dd_tx1[ind1_p + 1] - dd_tx1[ind1_p]) / (
            1 + np.abs(ddx1[ind1_p + 1]) / np.abs(ddx1[ind1_p])
        )
        t_zeros1_m = dd_tx1[ind1_m] + (dd_tx1[ind1_m + 1] - dd_tx1[ind1_m]) / (
            1 + np.abs(ddx1[ind1_m + 1]) / np.abs(ddx1[ind1_m])
        )

        ind2_p = ind_zeros2[ind_zeros2_p]
        ind2_m = ind_zeros2[ind_zeros2_m]

        t_zeros2_p = dd_tx2[ind2_p] + (dd_tx2[ind2_p + 1] - dd_tx2[ind2_p]) / (
            1 + np.abs(ddx2[ind2_p + 1]) / np.abs(ddx2[ind2_p])
        )
        t_zeros2_m = dd_tx2[ind2_m] + (dd_tx2[ind2_m + 1] - dd_tx2[ind2_m]) / (
            1 + np.abs(ddx2[ind2_m + 1]) / np.abs(ddx2[ind2_m])
        )

    # Define identical pairs of two time axis
    t1_dd_p, t2_dd_p, _, _ = find_closest(t_zeros1_p, t_zeros2_p)
    t1_dd_m, t2_dd_m, _, _ = find_closest(t_zeros1_m, t_zeros2_m)

    t1_dd = np.vstack([t1_dd_p, t1_dd_m])
    t1_dd = t1_dd[t1_dd[:, 0].argsort(), 0]

    t2_dd = np.vstack([t2_dd_p, t2_dd_m])
    t2_dd = t2_dd[t2_dd[:, 0].argsort(), 0]

    return t1_d, t2_d, t1_dd, t2_dd



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging

# Built-in imports
import os

from ..pyrf.dist_append import dist_append
from ..pyrf.ts_append import ts_append
from ..pyrf.ttns2datetime64 import ttns2datetime64
from .get_dist import get_dist
from .get_ts import get_ts
from .list_files import list_files

# Local imports
from .tokenize import tokenize

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)


def _var_and_cdf_name(var_str, mms_id):
    var = tokenize(var_str)

    root_path = os.path.dirname(os.path.abspath(__file__))

    with open(
        os.sep.join([root_path, "mms_keys.json"]), "r", encoding="utf-8"
    ) as json_file:
        keys_ = json.load(json_file)

    var["dtype"] = keys_[var["inst"]][var_str.lower()]["dtype"]
    cdf_name = f"mms{mms_id}_{keys_[var['inst']][var_str.lower()]['cdf_name']}"

    return var, cdf_name


def _check_times(inp):
    if inp.time.data.dtype == "int64":
        out = inp.assign_coords(time=ttns2datetime64(inp.time.data))
    else:
        out = inp
    return out


def get_data(var_str, tint, mms_id, verbose: bool = True, data_path: str = ""):
    r"""Load a variable. var_str must be in var (see below)

    Parameters
    ----------
    var_str : str
        Key of the target variable (use mms.get_data() to see keys.).
    tint : list of str
        Time interval.
    mms_id : str or int
        Index of the target spacecraft.
    verbose : bool, Optional
        Set to True to follow the loading. Default is True.
    data_path : str, Optional
        Path of MMS data. If None use `pyrfu/mms/config.json`

    Returns
    -------
    out : xarray.DataArray or xarray.Dataset
        Time series of the target variable of measured by the target
        spacecraft over the selected time interval.

    See also
    --------
    pyrfu.mms.get_ts : Read time series.
    pyrfu.mms.get_dist : Read velocity distribution function.

    Examples
    --------
    >>> from pyrfu import mms

    Define time interval

    >>> tint_brst = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Index of MMS spacecraft

    >>> ic = 1

    Load magnetic field from FGM

    >>> b_xyz = mms.get_data("b_gse_fgm_brst_l2", tint_brst, ic)

    """

    mms_id = str(mms_id)

    var, cdf_name = _var_and_cdf_name(var_str, mms_id)

    files = list_files(tint, mms_id, var, data_path)

    assert files, "No files found. Make sure that the data_path is correct"

    if verbose:
        logging.info("Loading %s...", cdf_name)

    out = None

    for file in files:
        if "-dist" in var["dtype"]:
            out = dist_append(out, get_dist(file, cdf_name, tint))

        else:
            out = ts_append(out, get_ts(file, cdf_name, tint))

    out = _check_times(out)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import signal

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


# noinspection PyTupleAssignmentBalance
def _ellip_coefficients(f_min, f_max, order):
    num1, den1, num2, den2 = [None] * 4

    if f_min == 0:
        if order == -1:
            order, f_max = signal.ellipord(
                f_max,
                np.min([f_max * 1.1, 0.9999]),
                0.5,
                60,
            )

        num1, den1 = signal.ellip(order, 0.5, 60, f_max, btype="lowpass")
    elif f_max == 0:
        if order == -1:
            order, f_min = signal.ellipord(
                f_min,
                np.min([f_min * 1.1, 0.9999]),
                0.5,
                60,
            )

        num1, den1 = signal.ellip(order, 0.5, 60, f_min, btype="highpass")
    else:
        if order == -1:
            order1, f_max = signal.ellipord(
                f_max,
                np.min([f_max * 1.3, 0.9999]),
                0.5,
                60,
            )
            order2, f_min = signal.ellipord(f_min, f_min * 0.75, 0.5, 60)
        else:
            order1, order2 = [order, order]

        num1, den1 = signal.ellip(order1, 0.5, 60, f_max)
        num2, den2 = signal.ellip(order2, 0.5, 60, f_min)

    return num1, den1, num2, den2


def filt(inp, f_min: float = 0.0, f_max: float = 1.0, order: int = -1):
    r"""Filters input quantity.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the variable to filter.
    f_min : float, Optional
        Lower limit of the frequency range. Default is 0. (Highpass filter).
    f_max : float, Optional
        Upper limit of the frequency range. Default is 1. (Highpass filter).
    order : int, Optional
        Order of the elliptic filter. Default is -1.

    Returns
    -------
    out : xarray.DataArray
        Time series of the filtered signal.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2017-07-18T13:03:34.000", "2017-07-18T13:07:00.000"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic and electric fields

    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_brst_l2", tint, mms_id)

    Convert E to field aligned coordinates

    >>> e_xyzfac = pyrf.convert_fac(e_xyz, b_xyz, [1,0,0])

    Bandpass filter E waveform

    >>> e_xyzfac_hf = pyrf.filt(e_xyzfac, 4, 0, 3)
    >>> e_xyzfac_lf = pyrf.filt(e_xyzfac, 0, 4, 3)

    """

    assert isinstance(inp, xr.DataArray), "inp must be a xarray.DataArray"

    f_samp = 1 / (np.median(np.diff(inp.time)).astype(np.int64) * 1e-9)

    # Data of the input
    inp_data = inp.data

    assert isinstance(f_min, (int, float)), "f_min must be int or float"
    assert isinstance(f_max, (int, float)), "f_max must be int or float"
    assert isinstance(order, (int, float)), "order must be int or float"

    f_min, f_max = [f_min / (f_samp / 2), f_max / (f_samp / 2)]

    f_max = np.min([f_max, 1.0])

    # Parameters of the elliptic filter. fact defines the width between
    # stopband and passband
    # r_pass, r_stop, fact = [0.5, 60, 1.1]

    num1, den1, num2, den2 = _ellip_coefficients(f_min, f_max, order)

    if len(inp_data.shape) == 1:
        inp_data = inp_data[:, np.newaxis]

    out_data = np.zeros(inp_data.shape)

    for i_col in range(inp_data.shape[1]):
        out_data[:, i_col] = signal.filtfilt(num1, den1, inp_data[:, i_col])

        if num2 is not None and den2 is not None:
            out_data[:, i_col] = signal.filtfilt(
                num2,
                den2,
                out_data[:, i_col],
            )
    if inp_data.shape[1] == 1:
        out_data = out_data[:, 0]

    out = xr.DataArray(
        out_data,
        coords=inp.coords,
        dims=inp.dims,
        attrs=inp.attrs,
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from ..mms.rotate_tensor import rotate_tensor
from .avg_4sc import avg_4sc
from .c_4_grad import c_4_grad
from .trace import trace
from .ts_scalar import ts_scalar
from .ts_tensor_xyz import ts_tensor_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def pid_4sc(r_mms, v_mms, p_mms, b_mms):
    r"""Compute Pi-D term using definition of [1]_ as :

    .. math::

        Pi-D = - \Pi_{ij}D_{ij}

    with :math:`\Pi_{ij}` the deviatoric part of the pressure
    tensor :

    .. math::

        \Pi_{ij} = P_{ij} - p\delta_{ij}

        p = \frac{1}{3}P_{ii}


    and :math:`D_{ij}` the deviatoric part of the strain tensor :

    .. math::

        D_{ij} = \frac{1}{2}\left ( \partial_i u_j + \partial_j u_i \right )
        - \frac{1}{3}\theta\delta_{ij}

        \theta = \nabla . u


    Parameters
    ----------
    r_mms : list of xarray.DataArray
        Time series of the position of the 4 spacecraft.
    v_mms : list of xarray.DataArray
        Time series of the bulk velocities of the 4 spacecraft.
    p_mms : list of xarray.DataArray
        Time series of the pressure tensor of the 4 spacecraft.
    b_mms : list of xarray.DataArray
        Time series of the background magnetic field of the 4 spacecraft.

    Returns
    -------
    pid : xarray.DataArray
        Time series of the Pi-D.

    References
    ----------
    .. [1]  Yang, Y., Matthaeus, W. H., Parashar, T. N., Wu, P., Wan, M.,
            Shi, Y., et al. (2017). Energy transfer channels and turbulence
            cascade in Vlasov-Maxwell turbulence. Physical Review E, 95,
            061201. doi : https://doi.org/10.1103/PhysRevE.95.061201

    """

    # Compute pressure tensor and background magnetic field at the center of
    # mass of the tetrahedron
    p_xyz = avg_4sc(p_mms)
    b_xyz = avg_4sc(b_mms)

    # Compute divergence and gradient tensor of the bulk velocity. Yang's
    # notation
    theta = c_4_grad(r_mms, v_mms, "div")
    grad_u = c_4_grad(r_mms, v_mms, "grad")

    # Define identity tensor
    identity_3d = np.zeros((len(grad_u), 3, 3))
    np.einsum("jii->ji", identity_3d)[:] = 1

    # strain tensor
    eps_xyz = (grad_u.data + np.transpose(grad_u.data, [0, 2, 1])) / 2

    # Deviatoric part of the strain tensor
    d_xyz = eps_xyz - theta.data[:, np.newaxis, np.newaxis] * identity_3d / 2
    d_xyz = ts_tensor_xyz(v_mms[0].time.data, d_xyz)

    # Convert tensors to field aligned coordinates
    d_fac = rotate_tensor(d_xyz, "fac", b_xyz, "pp")
    p_fac = rotate_tensor(p_xyz, "fac", b_xyz, "pp")

    # Isotropic scalar pressure
    press = trace(p_fac) / 3

    # Deviatoric part of the pressure tensor
    pi_fac = p_fac.data - press.data[:, np.newaxis, np.newaxis] * identity_3d
    pi_fac = ts_tensor_xyz(v_mms[0].time.data, pi_fac)

    # Compute Pi-D
    pid = np.sum(np.sum(pi_fac.data * d_fac.data, axis=1), axis=1)
    pid = ts_scalar(v_mms[0].time.data, pid)

    # Flatten tensors
    d_flat = np.reshape(d_fac.data, [len(d_fac), 9])
    pi_flat = np.reshape(pi_fac.data, [len(pi_fac), 9])
    d_flat = d_flat[:, [0, 4, -1, 1, 2, 5]]
    pi_flat = pi_flat[:, [0, 4, -1, 1, 2, 5]]

    # Compute components of the double contraction sum
    d_coeff = xr.DataArray(
        d_flat * pi_flat,
        coords=[p_xyz.time.data, np.arange(1, 7)],
        dims=["time", "index"],
    )

    return pid, d_coeff



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import xarray as xr
from scipy import signal

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def lowpass(inp, f_cut, fhz):
    r"""Filter the data through low or highpass filter with max
    frequency f_cut and subtract from the original.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input variable.
    f_cut : float
        Cutoff frequency.
    fhz : float
        Sampling frequency.

    Returns
    -------
    out : xarray.DataArray
        Time series of the filter data.

    """

    data = inp.data

    # Remove trend
    data_detrend = signal.detrend(data, type="linear", axis=0)
    rest = data - data_detrend

    # Elliptic filter
    f_nyq, r_pass, r_stop, order = [fhz / 2, 0.1, 60, 4]

    elliptic_filter = signal.ellip(
        order,
        r_pass,
        r_stop,
        f_cut / f_nyq,
        output="ba",
    )

    # Filter data
    out_data = signal.filtfilt(
        elliptic_filter[0],
        elliptic_filter[1],
        data_detrend,
        axis=0,
    )

    out = xr.DataArray(out_data + rest, coords=inp.coords, dims=inp.dims)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from ..pyrf.calc_fs import calc_fs
from ..pyrf.resample import resample
from ..pyrf.ts_tensor_xyz import ts_tensor_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def rotate_tensor(inp, rot_flag, vec, perp: str = "pp"):
    """Rotates pressure or temperature tensor into another coordinate system.

    Parameters
    ----------
    PeIJ/Peall : xarray.DataArray
        Time series of either separated terms of the tensor or the complete
        tensor. If columns (PeXX,PeXY,PeXZ,PeYY,PeYZ,PeZZ)

    rot_flag : str
        Flag of the target coordinates system :
            * "fac" : 	Field-aligned coordinates, requires background
                        magnetic field Bback, optional flag "pp"
                        :math:`\\mathbf{P}_{\\perp 1} = \\mathbf{P}_{\\perp2}`
                        or "qq" :math:`\\mathbf{P}_{\\perp 1}` and
                        :math:`\\mathbf{P}_{\\perp 2}` are most unequal, sets
                        P_{23} to zero.

            * "rot" :	Arbitrary coordinate system, requires new x-direction
                        xnew, new y and z directions ynew, znew (if not
                        included y and z directions are orthogonal to xnew
                        and closest to the original y and z directions)

            * "gse" : GSE coordinates, requires MMS spacecraft number

    vec : xarray.DataArray or numpy.ndarray
        Vector or coordinates system to rotate the tensor. If vec is timeseries of a
        vector tensor is rotated in field aligned coordinates. If vec is a
        numpy.ndarray rotates to a time independant coordinates system.

    perp : str, Optional
        Flag for perpandicular components of the tensor. Default is pp.
            * "pp" : perpendicular diagonal components are equal
            * "qq" : perpendicular diagonal components are most unequal


    Returns
    -------
    Pe : xarray.DataArray
        Time series of the pressure or temperature tensor in field-aligned,
        user-defined, or GSE coordinates.
        For "fac" Pe = [Ppar P12 P13; P12 Pperp1 P23; P13 P23 Pperp2].
        For "rot" and "gse" Pe = [Pxx Pxy Pxz; Pxy Pyy Pyz; Pxz Pyz Pzz]

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field and ion temperature tensor
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> t_xyz_i = mms.get_data("Ti_gse_fpi_fast_l2", tint, mms_id)
    >>> # Compute ion temperature in field aligned coordinates
    >>> t_xyzfac_i = mms.rotate_tensor(t_xyz_i, "fac", b_xyz, "pp")

    TODO : implement method "gse"
    """

    assert isinstance(rot_flag, str), "flag must be a string"
    assert rot_flag.lower() in ["fac", "rot", "gse"], "flag must be fac, rot or gse"

    assert isinstance(perp, str), "perp must be a string"
    assert perp.lower() in ["pp", "qq"], "perp must be pp or qq"

    # Check input and load pressure/temperature terms
    inp_times = inp.time.data
    n_t = len(inp_times)

    rot_mat = np.zeros((n_t, 3, 3))

    if rot_flag[0] == "f":
        assert isinstance(vec, xr.DataArray)
        b_back = resample(vec, inp, f_s=calc_fs(inp))

        b_vec = b_back / np.linalg.norm(b_back, axis=1, keepdims=True)

        r_x = b_vec.data
        r_y = np.array([1, 0, 0])
        r_z = np.cross(r_x, r_y)
        r_z /= np.linalg.norm(r_z, axis=1, keepdims=True)
        r_y = np.cross(r_z, r_x)
        r_y /= np.linalg.norm(r_y, axis=1, keepdims=True)

        rot_mat[:, 0, :], rot_mat[:, 1, :], rot_mat[:, 2, :] = [r_x, r_y, r_z]

    elif rot_flag[0] == "r":
        assert isinstance(vec, np.ndarray)

        if vec.ndim == 1 and vec.shape[0] == 3:
            r_x = vec
            r_x /= np.linalg.norm(r_x, keepdims=True)
            r_y = np.array([0, 1, 0])
            r_z = np.cross(r_x, r_y)
            r_z /= np.linalg.norm(r_z, keepdims=True)
            r_y = np.cross(r_z, r_x)
            r_y /= np.linalg.norm(r_y, keepdims=True)

        elif vec.ndim == 2 and vec.shape[0] == 3 and vec.shape[1] == 3:
            r_x = vec[:, 0] / np.linalg.norm(vec[:, 0], keepdims=True)
            r_y = vec[:, 1] / np.linalg.norm(vec[:, 1], keepdims=True)
            r_z = vec[:, 2] / np.linalg.norm(vec[:, 2], keepdims=True)

        else:
            raise TypeError("Vector format not recognized.")

        rot_mat[:, 0, :] = np.ones((n_t, 1)) * r_x
        rot_mat[:, 1, :] = np.ones((n_t, 1)) * r_y
        rot_mat[:, 2, :] = np.ones((n_t, 1)) * r_z

    else:
        raise NotImplementedError("gse is not yet implemented!!")

    p_tensor_p = np.zeros((n_t, 3, 3))

    for i in range(n_t):
        rot_temp = np.squeeze(rot_mat[i, :, :])

        p_tensor_p[i, :, :] = np.matmul(
            np.matmul(rot_temp, np.squeeze(inp.data[i, :, :])),
            np.transpose(rot_temp),
        )

    if perp.lower() == "pp":
        thetas = 0.5 * np.arctan(
            (p_tensor_p[:, 2, 2] - p_tensor_p[:, 1, 1]) / (2 * p_tensor_p[:, 1, 2]),
        )
        thetas[np.isnan(thetas)] = 0.0

        for i, theta in enumerate(thetas):
            rot_temp = np.array(
                [
                    [1, 0, 0],
                    [0, np.cos(theta), np.sin(theta)],
                    [0, -np.sin(theta), np.cos(theta)],
                ],
            )

            p_tensor_p[i, :, :] = np.matmul(
                np.matmul(rot_temp, np.squeeze(p_tensor_p[i, :, :])),
                np.transpose(rot_temp),
            )

    else:
        thetas = 0.5 * np.arctan(
            (2 * p_tensor_p[:, 1, 2]) / (p_tensor_p[:, 2, 2] - p_tensor_p[:, 1, 1]),
        )

        for i, theta in enumerate(thetas):
            rot_temp = np.array(
                [
                    [1, 0, 0],
                    [0, np.cos(theta), -np.sin(theta)],
                    [0, np.sin(theta), np.cos(theta)],
                ],
            )

            p_tensor_p[i, :, :] = np.matmul(
                np.matmul(rot_temp, np.squeeze(p_tensor_p[i, :, :])),
                np.transpose(rot_temp),
            )

    # Construct output
    p_new = ts_tensor_xyz(inp_times, p_tensor_p, attrs=inp.attrs)

    return p_new



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party
import numpy as np

# Local imports
from ..pyrf.calc_dt import calc_dt

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def psd_rebin(vdf, phi, energy0, energy1, esteptable):
    r"""Converts burst mode distribution into 64 energy channel distribution.
    Functions takes the burst mode distribution sampled in two energy tables
    and converts to a single energy table with 64 energy channels. Time
    resolution is halved and phi angles are averaged over adjacent times.

    Parameters
    ----------
    vdf : xarray.Dataset
        Time series of the particle distribution.
    phi : xarray.DataArray
        Time series of the phi angles.
    energy0 : numpy.ndarray
        Energy table 0.
    energy1 : numpy.ndarray
        Energy table 1.
    esteptable : numpy.ndarray
        Time series of the stepping table between energies (burst).

    Returns
    -------
    time_r : numpy.ndarray
        Revised time steps.
    vdf_r : numpy.ndarray
        Rebinned particle distribution.
    energy_r : numpy.ndarray
        Revised energy table.
    phi_r : numpy.ndarray
        Time series of the recalculated phi angle.

    Notes
    -----
    I'm assuming no gaps in the burst data interval. If there is a gap use
    time_clip before running. To be updated later.

    """

    assert isinstance(energy0, np.ndarray), "energy0 must be a numpy.ndarray"
    assert isinstance(energy1, np.ndarray), "energy1 must be a numpy.ndarray"
    assert isinstance(esteptable, np.ndarray), "step_table must be a numpy.ndarray"

    # Sort energy levels
    energy_r = np.sort(np.hstack([energy0, energy1]))

    # Define new times
    delta_t = calc_dt(vdf.data)
    time_r = vdf.time.data[:-1:2] + int(delta_t * 1e9 / 2)

    vdf_r = np.zeros((len(time_r), 64, *vdf.data.shape[2:]))
    phi_r = np.zeros((len(time_r), vdf.data.shape[2]))

    phi_s = np.roll(phi.data, 2, axis=1)
    phi_s[:, 0] = phi_s[:, 0] - 360

    time_indices = np.arange(0, len(vdf.time) - 1, 2)

    for new_el_num, idx in enumerate(time_indices[:-1]):
        if phi.data[idx, 0] > phi.data[idx + 1, 0]:
            phi_r[new_el_num, :] = (phi.data[idx, :] + phi_s[idx + 1, :]) / 2

            vdf_temp = np.roll(
                np.squeeze(vdf.data.data[idx + 1, ...]),
                2,
                axis=1,
            )

            if esteptable[idx]:
                vdf_r[new_el_num, 1:64:2, ...] = vdf.data.data[idx, ...]
                vdf_r[new_el_num, 0:63:2, ...] = vdf_temp
            else:
                vdf_r[new_el_num, 1:64:2, ...] = vdf.data.data[idx, ...]
                vdf_r[new_el_num, 0:63:2, ...] = vdf_temp

        else:
            phi_r[new_el_num, :] = phi.data[idx, :] + phi.data[idx + 1, :]
            phi_r[new_el_num, :] /= 2

            if esteptable[idx]:
                vdf_r[new_el_num, 1:64:2, ...] = vdf.data.data[idx, ...]
                vdf_r[new_el_num, 0:63:2, ...] = vdf.data.data[idx + 1, ...]
            else:
                vdf_r[new_el_num, 1:64:2, ...] = vdf.data.data[idx + 1, ...]
                vdf_r[new_el_num, 0:63:2, ...] = vdf.data.data[idx, ...]

    return time_r, vdf_r, energy_r, phi_r



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def ts_append(inp0, inp1):
    r"""Concatenate two time series along the time axis.

    Parameters
    ----------
    inp0 : xarray.DataArray
        Time series of the first input (early times).
    inp1 : xarray.DataArray
        Time series of the second input (late times).

    Returns
    -------
    out : xarray.DataArray
        Concatenated time series.

    Notes
    -----
    The time series must be in the correct time order.

    """

    if inp0 is None:
        return inp1

    out_data = {}

    if inp0.data.ndim != 1:
        out_data["data"] = np.vstack([inp0, inp1])

    else:
        out_data["data"] = np.hstack([inp0, inp1])

    out_data["attrs"] = {}

    for k in inp0.attrs:
        if isinstance(inp0.attrs[k], np.ndarray):
            out_data["attrs"][k] = np.hstack([inp0.attrs[k], inp1.attrs[k]])

        else:
            out_data["attrs"][k] = inp0.attrs[k]

    depends = [{} for _ in range(len(inp0.dims))]

    for i, dim in enumerate(inp0.dims):
        if i == 0 or dim == "time":
            depends[i]["data"] = np.hstack([inp0[dim].data, inp1[dim].data])

            # add attributes
            depends[i]["attrs"] = {}

            for k in inp0[dim].attrs:
                # if attrs is array time append
                if isinstance(inp0[dim].attrs[k], np.ndarray):
                    depends[i]["attrs"][k] = np.hstack(
                        [inp0[dim].attrs[k], inp1[dim].attrs[k]],
                    )

                else:
                    depends[i]["attrs"][k] = inp0[dim].attrs[k]

        else:
            # Use values of other coordinates of inp0 assuming equal to inp1
            depends[i]["data"] = inp0[dim].data

            # add attributes
            depends[i]["attrs"] = {}

            for k in inp0[dim].attrs:
                depends[i]["attrs"][k] = inp0[dim].attrs[k]

    # Create DataArray
    out = xr.DataArray(
        out_data["data"],
        coords=[depend["data"] for depend in depends],
        dims=inp0.dims,
        attrs=out_data["attrs"],
    )

    for i, dim in enumerate(out.dims):
        out[dim].attrs = depends[i]["attrs"]

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def calc_fs(inp):
    r"""Computes the sampling frequency of the input time series.

    Parameters
    ----------
    inp : xarray.DataArray or xarray.Dataset
        Time series of the input variable.

    Returns
    -------
    out : float
        Sampling frequency in Hz.

    """

    message = "Input must be a time series"
    assert isinstance(inp, (xr.Dataset, xr.DataArray)), message

    out = 1 / (np.median(np.diff(inp.time.data)).astype(np.float64) * 1e-9)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np

# 3rd party imports
import xarray as xr
from scipy import signal

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def medfilt(inp, n_pts: int = 11):
    r"""Applies a median filter over npts points to inp.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input variable.
    n_pts : int, Optional
        Number of points of median filter.

    Returns
    -------
    out : xarray.DataArray
        Time series of the median filtered input variable.

    Examples
    --------
    >>> import numpy
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Spacecraft indices

    >>> mms_list = numpy.arange(1,5)

    Load magnetic field and electric field

    >>> r_mms, b_mms = [[] * 4 for _ in range(2)]
    >>> for mms_id in range(1, 5):
    >>> 	r_mms.append(mms.get_data("R_gse", tint, mms_id))
    >>> 	b_mms.append(mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id))
    >>>

    Compute current density, etc

    >>> j_xyz, _, b_xyz, _, _, _ = pyrf.c_4_j(r_mms, b_mms)

    Get J sampling frequency

    >>> fs = pyrf.calc_fs(j_xyz)

    Median filter over 1s

    >>> j_xyz = pyrf.medfilt(j_xyz,fs)

    """

    if isinstance(n_pts, float):
        n_pts = np.floor(n_pts).astype(np.int64)

    if n_pts % 2 == 0:
        n_pts += 1

    n_times = len(inp)

    if inp.ndim == 3:
        inp_data = np.reshape(inp.data, [n_times, 9])
    else:
        inp_data = inp.data

    try:
        n_comp = inp_data.shape[1]
    except IndexError:
        n_comp = 1

        inp_data = inp_data[..., None]

    out_data = np.zeros(inp_data.shape)

    if not n_pts % 2:
        n_pts += 1

    for i in range(n_comp):
        out_data[:, i] = signal.medfilt(inp_data[:, i], n_pts)

    if n_comp == 9:
        out_data = np.reshape(out_data, [n_times, 3, 3])

    if out_data.shape[1] == 1:
        out_data = out_data[:, 0]

    out = xr.DataArray(out_data, coords=inp.coords, dims=inp.dims)
    out.attrs = inp.attrs

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime

# Built-in imports
import urllib

# 3rd party imports
import numpy as np
import pandas as pd

# Local imports
from .iso86012datetime64 import iso86012datetime64

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


var_omni_1 = {
    "b": 13,
    "avgb": -1,
    "blat": -1,
    "blong": -1,
    "bx": 14,
    "bxgse": 14,
    "bxgsm": 14,
    "by": 15,
    "bygse": 15,
    "bz": 16,
    "bzgse": 16,
    "bygsm": 17,
    "bzgsm": 18,
    "t": 26,
    "n": 25,
    "nanp": -1,
    "v": 21,
    "vx": 22,
    "vy": 23,
    "vz": 24,
    "vlon": -1,
    "vlat": -1,
    "p": 27,
    "e": 28,
    "beta": 29,
    "ma": 30,
    "bsnx": 34,
    "bsny": 35,
    "bsnz": 36,
    "ms": 45,
    "ssn": -1,
    "dst": -1,
    "ae": 37,
    "al": 38,
    "au": 39,
    "kp": -1,
    "pc": 44,
    "f10.7": -1,
    "imfid": 4,
    "swid": 5,
    "ts": 9,
    "rmsts": 10,
}

var_omni_2 = {
    "b": 8,
    "avgb": 9,
    "blat": 10,
    "blong": 11,
    "bx": 12,
    "bxgse": 12,
    "bxgsm": 12,
    "by": 13,
    "bygse": 13,
    "bz": 14,
    "bzgse": 14,
    "bygsm": 15,
    "bzgsm": 16,
    "t": 22,
    "n": 23,
    "nanp": 27,
    "v": 24,
    "vx": -1,
    "vy": -1,
    "vz": -1,
    "vlon": 25,
    "vlat": 26,
    "p": 28,
    "e": 35,
    "beta": 36,
    "ma": 37,
    "bsnx": -1,
    "bsny": -1,
    "bsnz": -1,
    "ms": 56,
    "ssn": 39,
    "dst": 40,
    "ae": 41,
    "al": 52,
    "au": 53,
    "kp": 38,
    "pc": 51,
    "f10.7": 50,
    "imfid": -1,
    "swid": -1,
    "ts": -1,
    "rmsts": -1,
}


def _omni_url(tint, omni_database):
    if omni_database == "omni_hour":
        data_source = "omni2"
        date_format = "%Y%m%d"
        delta_t_min = 24 * 3600
    elif omni_database == "omni_min":
        data_source = "omni_min"
        date_format = "%Y%m%d%H"
        delta_t_min = 3600
    else:
        raise ValueError("Invalid database")

    url_ = "omniweb.gsfc.nasa.gov/cgi/nx1.cgi?activity=retrieve&spacecraft="
    url_ = f"https://{url_}{data_source}"

    tint[0] += np.timedelta64(0, "[s]")
    tint[1] += np.timedelta64(delta_t_min, "[s]")
    tint = [t_.astype(datetime.datetime) for t_ in tint]
    start_date, end_date = [t_.strftime(date_format) for t_ in tint]

    url_ = f"{url_}&start_date={start_date}&end_date={end_date}"

    return url_


def get_omni_data(variables, tint, database: str = "omni_hour"):
    r"""Downloads OMNI data.

    Parameters
    ----------
    variables : list
        Keys of the variables to download.
    tint : list
        Time interval.
    database : {"omni_hour", "omni_min"}, Optional
        OMNI data resolution. Default is database = "omni_hour".

    Returns
    -------
    data : xarray.Dataset
        OMNI data.

    """

    tint = iso86012datetime64(np.array(tint)).astype("datetime64[s]")

    url_ = _omni_url(tint, database)

    vars_ = ""
    for variable in variables:
        vars_ = f"{vars_}&vars={var_omni_2[variable]:d}"

    with urllib.request.urlopen(f"{url_}{vars_}") as file:
        out = str(file.read())

    idx_start, idx_end = [out.find("YEAR"), out.find("</pre>")]

    lines = out[idx_start:idx_end].split("\\n")[:-1]
    lines = [list(filter(lambda x: x != "", l_.split(" "))) for l_ in lines]
    lines = [[f"{l_[0]}-{l_[1]}/{l_[2]}", *l_[3:]] for l_ in lines[1:]]
    data = pd.DataFrame(lines, columns=["time", *variables])
    data["time"] = pd.to_datetime(data["time"], format="%Y-%j/%H")

    data = data.set_index("time").astype(np.float64)
    fmt_ = f"datetime64[{database[5].lower()}]"
    data = data.loc[data.index.isin(tint.astype(fmt_).astype("datetime64[ns]"))]
    data = data.to_xarray()

    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import logging
from typing import Union

# 3rd party imports
import numpy as np
from scipy import interpolate

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

logging.captureWarnings(True)
logging.basicConfig(
    format="[%(asctime)s] %(levelname)s: %(message)s",
    datefmt="%d-%b-%y %H:%M:%S",
    level=logging.INFO,
)

surface_materials = [
    "cluster",
    "themis",
    "cassini",
    "aluminium",
    "aquadag",
    "gold",
    "graphite",
    "solar cells",
    "1eV",
    "TiN",
    "elgiloy",
]

j_zeros = {
    "cassini": 25e-6,
    "tin": 25e-6,
    "cluster": 25e-6,
    "aluminium": 30e-6,
    "aquadag": 18e-6,
    "gold": 29e-6,
    "graphite": 7.2e-6,
    "solar cells": 20e-6,
    "solar cell": 20e-6,
    "elgiloy": 30e-6,
}


def photo_current(
    iluminated_area: float = None,
    u: Union[float, np.ndarray] = None,
    distance_sun: float = None,
    flag: Union[str, float] = "cluster",
) -> Union[float, np.ndarray]:
    r"""Calculates the photo-current emitted by an arbitrary body.

    Parameters
    ----------
    iluminated_area : float
        Cross section area [m^2].
    u : float or numpy.ndarray
        Potential [V].
    distance_sun : float
        Distance form the Sun [AU].
    flag : str or float, Optional
        Surface materials or surface photoemission in [A/m^2].
        Default is "cluster".

    Returns
    -------
    j_photo : float or numpy.ndarray
        Photo-current emitted.

    Notes
    -----
    Estimates are done for the solar minimum conditions.

    """

    assert isinstance(flag, (str, float))

    if not iluminated_area and not u and not distance_sun:
        for surf in surface_materials:
            j0 = photo_current(1, 0, 1, surf)
            logging.info(
                "%(surf)s: Io= %(i0)3.2f uA/m2", {"surf": surf, "i0": j0 * 1e6}
            )

        return None

    # Assert than u is an array
    u = np.atleast_1d(u)

    if isinstance(flag, (float, int)):
        photoemisson = flag

        # Initialize
        j_photo = np.ones(u.shape)

        # initialize to current valid for negative potentials
        j_photo *= photoemisson * iluminated_area / distance_sun**2

        a_ = 5.0e-5 / 5.6e-5 * np.exp(-u[u >= 0.0] / 2.74)
        b_ = 1.2e-5 / 5.6e-5 * np.exp(-(u[u >= 0] + 10.0) / 14.427)
        j_photo[u >= 0] *= a_ + b_

    elif flag.lower() == "1ev":
        j_photo = np.ones(u.shape)

        # initialize to current valid for negative potentials
        j_photo *= 5.0e-5 * iluminated_area / distance_sun**2

        j_photo[u >= 0] *= np.exp(-u[u >= 0])

    elif flag.lower() == "themis":
        ref_u = np.array([0.1, 1, 5, 10, 50])
        ref_j_photo = np.array([50, 27, 10, 5, 0.5]) * 1e-6
        log_u = np.log(ref_u)
        log_j = np.log(ref_j_photo)

        # Initialize
        j_photo = np.ones(u.shape)

        # initialize to current valid for negative potentials
        j_photo *= iluminated_area / distance_sun**2

        f_ = interpolate.PchipInterpolator(log_u, log_j, extrapolate=None)
        j_photo[u >= ref_u[0]] *= np.exp(f_(np.log(u[u >= ref_u[0]])))
        j_photo[u < ref_u[0]] *= ref_j_photo[0]

    elif flag.lower() in j_zeros:
        j_photo = photo_current(iluminated_area, u, distance_sun, "themis")
        j_photo *= j_zeros[flag] / photo_current(1.0, 0.0, 1.0, "themis")

    else:
        raise ValueError("Unknown surface material.")

    return j_photo



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy import signal

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def wave_fft(
    inp,
    window,
    frame_overlap: float = 10.0,
    frame_length: float = 20.0,
    f_sampling: float = None,
):
    r"""Short-Time Fourier Transform.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the one dimension data.
    window : str
        Window function such as rectwin, hamming (default).
    frame_overlap : float, Optional
        Length of each frame overlaps in second.
    frame_length : float, Optional
        Length of each frame in second.
    f_sampling : float, Optional
        Sampling frequency.

    Returns
    -------
    spectrogram : ndarray
        Spectrogram of x.
    time : ndarray
        Value corresponds to the center of each frame (x-axis) in sec.
    frequencies : ndarray
        Vector of frequencies (y-axis) in Hz.

    """

    if f_sampling is None:
        delta_t = np.median(np.diff(inp.time.data).astype(np.float64)) * 1e-9
        f_sampling = 1 / delta_t

    # convert ms to points
    n_per_seg = np.round(frame_length * f_sampling).astype(np.int64)
    n_overlap = np.round(frame_overlap * f_sampling).astype(np.int64)

    options = {
        "fs": f_sampling,
        "window": window,
        "nperseg": n_per_seg,
        "noverlap": n_overlap,
        "mode": "complex",
    }
    frequencies, time, spectrogram = signal.spectrogram(inp, **options)

    return frequencies, time, spectrogram



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

from ..pyrf.ts_skymap import ts_skymap

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _idx_closest(lst0, lst1):
    return [(np.abs(np.asarray(lst0) - k)).argmin() for k in lst1]


def _combine_attrs(skymaps_attrs):
    attrs = {}
    filtered_attrs = [
        "delta_energy_plus",
        "delta_energy_minus",
        "esteptable",
        "energy0",
        "energy1",
    ]

    attrs_keys = list(
        filter(lambda k: k not in filtered_attrs, skymaps_attrs[0]),
    )

    for k in attrs_keys:
        sms_attrs = [skymaps_attr[k] for skymaps_attr in skymaps_attrs]

        is_same = [sms_attr == sms_attrs[0] for sms_attr in sms_attrs[1:]]

        try:
            if all(is_same):
                attrs[k] = sms_attrs[0]
            else:
                continue
        except ValueError:
            continue

    return attrs


def eis_skymap_combine_sc(skymaps, method: str = "mean"):
    r"""Generate composite skymap from the EIS sensors across the MMS
    spacecraft.

    Parameters
    ----------
    skymaps : list of xarray.Dataset
        Skymap distribution for all spacecraft.
    method : str, Optional
        Method to combine spectra, "mean" or "sum"

    Returns
    -------
    out : xarray.Dataset
        Composite skymap distribution

    See Also
    --------
    pyrfu.mms.get_eis_allt, pyrfu.mms.eis_pad,
    pyrfu.mms.eis_spec_combine_sc, pyrfu.mms.eis_spec_combine_sc

    """
    assert method.lower() in ["mean", "sum"]

    # Determine spacecraft with smallest number of time steps to use as
    # reference spacecraft
    time_size = [len(probe.time.data) for probe in skymaps]
    ref_sc_time_size, ref_sc_loc = [np.min(time_size), np.argmin(time_size)]
    ref_probe = skymaps[ref_sc_loc]

    # Define common energy grid across EIS instruments
    n_en_chans = [probe.energy.shape[1] for probe in skymaps]
    size_en, loc_ref_en = [np.min(n_en_chans), np.argmin(n_en_chans)]
    ref_energy = skymaps[loc_ref_en].energy.data[0, :]

    energy_data, e_plus, e_minu = [[], [], []]
    for probe in skymaps:
        idx = _idx_closest(probe.energy.data[0, :], ref_energy)
        energy_data.append(probe.energy.data[0, idx])
        e_minu.append(probe.attrs["delta_energy_minus"][idx])
        e_plus.append(probe.attrs["delta_energy_plus"][idx])

    energy_data = np.stack(energy_data)
    common_energy = np.nanmean(energy_data, axis=0)
    common_energy = np.tile(common_energy, (ref_sc_time_size, 1))

    #
    e_minu = np.stack(e_minu)
    e_plus = np.stack(e_plus)
    common_minu = np.nanmean(e_minu, axis=0)
    common_plus = np.nanmean(e_plus, axis=0)

    # Use azimuthal and elevation angle from reference spacecraft (in
    # practice they are the same for all spacecraft)
    phi = ref_probe.phi.data
    theta = ref_probe.theta.data

    allmms_skymap = np.zeros(
        [ref_sc_time_size, size_en, phi.shape[1], len(theta), len(skymaps)],
    )

    for i_s, skymap in enumerate(skymaps):
        idx_en = _idx_closest(skymap.energy.data[0, :], common_energy[0, :])
        allmms_skymap[..., i_s] = skymap.data[:ref_sc_time_size, idx_en, ...]

    if method.lower() == "mean":
        # Average the four spacecraft
        allmms_skymap_avg = np.nanmean(allmms_skymap, axis=-1)
    else:
        # Sum the four spacecraft
        allmms_skymap_avg = np.nansum(allmms_skymap, axis=-1)

    # Combine attributes
    # Combine global attributes
    glob_attrs = _combine_attrs([s.attrs for s in skymaps])
    glob_attrs = {
        "delta_energy_plus": np.tile(common_plus, (ref_sc_time_size)),
        "delta_energy_minus": np.tile(common_minu, (ref_sc_time_size)),
        **glob_attrs,
    }

    # Combine coordinates attributes
    coords_attrs = {}
    for coor in ["time", "energy", "phi", "theta"]:
        coords_attrs[coor] = _combine_attrs(
            [skymap[coor].attrs for skymap in skymaps],
        )

    # Combine data attributes
    attrs = _combine_attrs([skymap.data.attrs for skymap in skymaps])

    # Create combined skymap
    out = ts_skymap(
        ref_probe.time.data,
        allmms_skymap_avg,
        common_energy,
        phi,
        theta,
        energy0=common_energy[0, :],
        energy1=common_energy[1, :],
        esteptable=np.zeros(len(ref_probe.time.data)),
        attrs=attrs,
        coords_attrs=coords_attrs,
        glob_attrs=glob_attrs,
    )

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr
from scipy import constants

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _mass_ratio(inp):
    if inp.attrs["species"].lower() in ["ions", "ion", "protons", "proton"]:
        mass_ratio = 1
    elif inp.attrs["species"].lower() in ["alphas", "alpha", "helium"]:
        mass_ratio = 4
    elif inp.attrs["species"].lower() in ["electrons", "e"]:
        mass_ratio = constants.electron_mass / constants.proton_mass
    else:
        raise ValueError("Invalid specie")

    return mass_ratio


def _convert(inp, units, mass_ratio):
    fact = 1 / (1e6 * 0.53707 * mass_ratio**2)

    if units.lower() == "s^3/cm^6":
        out = inp * 1e30 * fact
    elif units.lower() == "s^3/m^6":
        out = inp * 1e18 * fact
    elif units.lower() == "s^3/km^6":
        out = inp * fact
    else:
        raise ValueError("Invalid unit")

    return out


def psd2def(inp):
    r"""Changes units to differential energy flux.

    Parameters
    ----------
    vdf : xarray.Dataset
        Time series of the 3D velocity distribution with :
            * time : Time samples.
            * data : 3D velocity distribution.
            * energy : Energy levels.
            * phi : Azimuthal angles.
            * theta : Elevation angle.

    Returns
    -------
    out : xarray.Dataset
        Time series of the 3D differential energy flux with :
            * time : Time samples.
            * data : 3D density energy flux.
            * energy : Energy levels.
            * phi : Azimuthal angles.
            * theta : Elevation angle.

    """

    assert isinstance(inp, (xr.DataArray, xr.Dataset)), "inp must be a xarray"

    if isinstance(inp, xr.Dataset):
        tmp_data = _convert(inp.data.data, inp.data.attrs["UNITS"], _mass_ratio(inp))
        energy = inp.energy.data
        energy_mat = np.tile(energy[:, :, None, None], (1, 1, *tmp_data.shape[2:]))
        tmp_data *= energy_mat**2
        out = inp.copy()
        out.data.data = np.squeeze(tmp_data)
        out.data.attrs["UNITS"] = "keV/(cm^2 s sr keV)"
    else:
        tmp_data = _convert(inp.data, inp.attrs["UNITS"], _mass_ratio(inp))
        energy = inp.energy.data
        energy_mat = np.tile(energy, (tmp_data.shape[0], 1))
        tmp_data *= energy_mat**2
        out = inp.copy()
        out.data = np.squeeze(tmp_data)
        out.attrs["UNITS"] = "keV/(cm^2 s sr keV)"

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def histogram(inp, bins=100, y_range=None, weights=None, density=True):
    r"""Computes 1D histogram of the inp with bins bins

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input scalar variable.
    bins : str or int or array_like, Optional
        Number of bins. Default is ``bins=100``.
    y_range : (float, float), Optional
        The lower and upper range of the bins.  If not provided, range
        is simply ``(inp.min(), inp.max())``.  Values outside the range are
        ignored. The first element of the range must be less than or
        equal to the second. `range` affects the automatic bin
        computation as well. While bin width is computed to be optimal
        based on the actual data within `range`, the bin count will fill
        the entire range including portions containing no data.
    weights : array_like, Optional
        An array of weights, of the same shape as `inp`.  Each value in
        `inp` only contributes its associated weight towards the bin count
        (instead of 1). If `density` is True, the weights are
        normalized, so that the integral of the density over the range
        remains 1.
    density : bool, Optional
        If ``False``, the result will contain the number of samples in each
        bin. If ``True``, the result is the value of the probability *density*
        function at the bin, normalized such that the *integral* over the
        range is 1. Note that the sum of the histogram values will not be
        equal to 1 unless bins of unity width are chosen; it is not a
        probability mass function.

    Returns
    -------
    out : xarray.DataArray
        1D distribution of the input time series.

    """

    # Check input
    assert isinstance(inp, xr.DataArray), "inp must be a xarray.DataArray"
    assert inp.ndim == 1, "inp must be a scalar time series"

    hist, bins = np.histogram(
        inp.data,
        bins=bins,
        range=y_range,
        weights=weights,
        density=density,
    )
    bin_center = (bins[1:] + bins[:-1]) * 0.5

    out = xr.DataArray(hist, coords=[bin_center], dims=["bins"])

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import xarray as xr

from .db_get_ts import db_get_ts

# Local imports
from .feeps_active_eyes import feeps_active_eyes

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"

data_units_keys = {
    "flux": "intensity",
    "counts": "counts",
    "cps": "count_rate",
    "mask": "sector_mask",
}


def _tokenize(tar_var):
    var = {"inst": "feeps"}

    data_units = data_units_keys[tar_var.split("_")[0][:-1].lower()]

    specie = tar_var.split("_")[0][-1]

    if specie == "e":
        var["dtype"] = "electron"
    elif specie == "i":
        var["dtype"] = "ion"
    else:
        raise ValueError("invalid specie")

    var["tmmode"] = tar_var.split("_")[1]
    var["lev"] = tar_var.split("_")[2]

    return var, data_units


def _get_oneeye(
    tar_var,
    e_id,
    tint,
    mms_id,
    verbose: bool = True,
    data_path: str = "",
):
    mms_id = int(mms_id)

    var, data_units = _tokenize(tar_var)

    dset_name = f"mms{mms_id:d}_feeps_{var['tmmode']}_l2_{var['dtype']}"
    pref = f"epd_feeps_{var['tmmode']}_{var['lev']}_{var['dtype']}"

    active_eyes = feeps_active_eyes(var, tint, mms_id)

    if e_id.split("-")[0] in ["top", "bottom"]:
        suf = e_id.split("-")[0]
        e_id = int(e_id.split("-")[1])

        assert e_id in active_eyes[suf], "Unactive eye"

        suf = f"{suf}_{data_units}_sensorid_{e_id:d}"

    else:
        raise ValueError("Invalid format of eye id")

    out = db_get_ts(
        dset_name,
        f"mms{mms_id:d}_{pref}_{suf}",
        tint,
        verbose,
        data_path=data_path,
    )

    out.attrs["tmmode"] = var["tmmode"]
    out.attrs["lev"] = var["lev"]
    out.attrs["mms_id"] = mms_id
    out.attrs["dtype"] = var["dtype"]
    out.attrs["species"] = f"{var['dtype']}s"
    return out


def get_feeps_alleyes(
    tar_var,
    tint,
    mms_id,
    verbose: bool = True,
    data_path: str = "",
):
    r"""Read energy spectrum of the selected specie in the selected energy
    range for all FEEPS eyes.

    Parameters
    ----------
    tar_var : str
        Key of the target variable like
        {data_unit}{specie}_{data_rate}_{data_lvl}.
    tint : list of str
        Time interval.
    mms_id : int or float or str
        Index of the spacecraft.
    verbose : bool, Optional
        Set to True to follow the loading. Default is True.
    data_path : str, Optional
        Path of MMS data. Default uses `pyrfu.mms.mms_config.py`

    Returns
    -------
    out : xarray.Dataset
        Dataset containing the energy spectrum of the available eyes of the
        Fly's Eye Energetic Particle Spectrometer (FEEPS).

    Examples
    --------
    >>> from pyrfu import mms

    Define time interval

    >>> tint_brst = ["2017-07-23T16:54:24.000", "2017-07-23T17:00:00.000"]

    Read electron energy spectrum for all FEEPS eyes

    >>> feeps_all_eyes = mms.get_feeps_alleyes("fluxe_brst_l2", tint_brst, 2)

    """

    mms_id = int(mms_id)

    specie = tar_var.split("_")[0][-1]

    var = {
        "tmmode": tar_var.split("_")[1],
        "lev": tar_var.split("_")[2],
        "mmsId": mms_id,
    }

    if specie == "e":
        var["dtype"] = "electron"
    elif specie == "i":
        var["dtype"] = "ion"
    else:
        raise ValueError("Invalid specie")

    dset_name = f"mms{mms_id:d}_feeps_{var['tmmode']}_l2_{var['dtype']}"
    pref = f"epd_feeps_{var['tmmode']}_{var['lev']}_{var['dtype']}"

    active_eyes = feeps_active_eyes(var, tint, mms_id)

    e_ids = [f"{k}-{s:d}" for k in active_eyes for s in active_eyes[k]]

    out_dict = {
        "spinsectnum": db_get_ts(
            dset_name,
            f"mms{mms_id:d}_{pref}_spinsectnum",
            tint,
            data_path=data_path,
        ),
        "pitch_angle": db_get_ts(
            dset_name,
            f"mms{mms_id:d}_{pref}_pitch_angle",
            tint,
            data_path=data_path,
        ),
    }

    for e_id in e_ids:
        out_dict[e_id] = _get_oneeye(
            tar_var,
            e_id,
            tint,
            mms_id,
            verbose,
            data_path=data_path,
        )

        out_dict[e_id] = out_dict[e_id].rename(
            {out_dict[e_id].dims[1]: f"energy_{e_id}"}
        )

    out = xr.Dataset(out_dict)

    out.attrs = var

    out.attrs["specie"] = var["dtype"]

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
from scipy import constants

# Local imports
from ..mms import rotate_tensor
from .resample import resample
from .ts_scalar import ts_scalar

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def plasma_beta(b_xyz, p_xyz):
    """Computes plasma beta at magnetic field sampling

    .. math::

        \beta = \frac{P_{th}}{P_b}

    where : :math:`P_b = B^2 / 2 \\mu_0`

    Parameters
    ----------
    b_xyz : xarray.DataArray
        Time series of the magnetic field.
    p_xyz : xarray.DataArray
        Time series of the pressure tensor.

    Returns
    -------
    beta : xarray.DataArray
        Time series of the plasma beta at magnetic field sampling.

    """

    p_xyz = resample(p_xyz, b_xyz)

    p_fac = rotate_tensor(p_xyz, "fac", b_xyz, "pp")

    # Scalar temperature
    p_tot = (p_fac.data[:, 0, 0] + p_fac.data[:, 1, 1] + p_fac.data[:, 2, 2]) / 3

    # Magnitude of the magnetic field
    b_mag = np.linalg.norm(b_xyz.data, axis=1)
    p_mag = 1e-18 * b_mag**2 / (2 * constants.mu_0)

    # Compute plasma beta
    beta = ts_scalar(b_xyz.time.data, p_tot / p_mag)

    return beta



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
import itertools

# 3rd party imports
import numpy as np
import xarray as xr

from .avg_4sc import avg_4sc
from .c_4_k import c_4_k
from .cross import cross
from .dot import dot
from .normalize import normalize

# Local imports
from .resample import resample

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _to_ts(out_data, b_dict):
    if len(out_data.shape) == 1:
        out = xr.DataArray(out_data, coords=[b_dict["1"].time], dims=["time"])

    elif len(out_data.shape) == 2:
        out = xr.DataArray(
            out_data,
            coords=[b_dict["1"].time, ["x", "y", "z"]],
            dims=["time", "comp"],
        )

    else:
        out = xr.DataArray(
            out_data,
            coords=[b_dict["1"].time, ["x", "y", "z"], ["x", "y", "z"]],
            dims=["time", "vcomp", "hcomp"],
        )

    return out


def c_4_grad(r_list, b_list, method: str = "grad"):
    r"""Calculate gradient of physical field using 4 spacecraft technique
    in [2]_ [3]_.

    Parameters
    ----------
    r_list : list of xarray.DataArray
        Time series of the positions of the spacecraft
    b_list : list of xarray.DataArray
        Time series of the magnetic field at the corresponding positions
    method : {"grad", "div", "curl", "bdivb", "curv"}, Optional
        Method flag :
            * "grad" : compute gradient (default)
            * "div" : compute divergence
            * "curl" : compute curl
            * "bdivb" : compute b.div(b)
            * "curv" : compute curvature

    Returns
    -------
    out : xarray.DataArray
        Time series of the derivative of the input field corresponding to
        the method

    See also
    --------
    pyrfu.pyrf.c_4_k : Calculates reciprocal vectors in barycentric
                        coordinates.

    References
    ----------
    .. [2]	Dunlop, M. W., A. Balogh, K.-H. Glassmeier, and P. Robert (2002a),
            Four-point Cluster application of	magnetic field analysis
            tools: The Curl- ometer, J. Geophys. Res., 107(A11), 1384,
            doi : https://doi.org/10.1029/2001JA005088.

    .. [3]	Robert, P., et al. (1998), Accuracy of current determination, in
            Analysis Methods for Multi-Spacecraft Data, edited by G.
            Paschmann and P. W. Daly, pp. 395–418, Int. Space Sci. Inst.,
            Bern. doi : https://www.issibern.ch/forads/sr-001-16.pdf

    Examples
    --------
    >>> from pyrfu.mms import get_data
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]

    Load magnetic field and spacecraft position

    >>> b_mms = [get_data("B_gse_fgm_srvy_l2", tint, m) for m in range(1, 5)]
    >>> r_mms = [get_data("R_gse", tint, m) for m in range(1, 5)]
    >>> gradb = pyrf.c_4_grad(r_mms, b_mms, "grad")

    """

    assert isinstance(r_list, list) and len(r_list) == 4, "r_list must a list of s/c"
    assert isinstance(b_list, list) and len(b_list) == 4, "b_list must a list of s/c"

    assert isinstance(method, str), "method must be a string"
    assert method.lower() in ["grad", "div", "curl", "bdivb", "curv"], "Invalid method"

    # Resample with respect to 1st spacecraft
    r_list = [resample(r, b_list[0]) for r in r_list]
    b_list = [resample(b, b_list[0]) for b in b_list]

    # Compute reciprocal vectors in barycentric coordinates (see c_4_k)
    k_list = c_4_k(r_list)

    # Magnetic field at the center of mass of the tetrahedron
    b_avg = avg_4sc(b_list)

    b_dict = {"1": b_list[0], "2": b_list[1], "3": b_list[2], "4": b_list[3]}
    k_dict = {"1": k_list[0], "2": k_list[1], "3": k_list[2], "4": k_list[3]}

    mms_list = b_dict.keys()

    # Gradient of scalar/vector
    if len(b_dict["1"].shape) == 1:
        grad_b = np.zeros((len(b_dict["1"]), 3))

        for i_sc in mms_list:
            grad_b += k_dict[i_sc].data * np.tile(b_dict[i_sc].data, (3, 1)).T

    else:
        grad_b = np.zeros((len(b_dict["1"]), 3, 3))

        for i, j, i_sc in itertools.product(range(3), range(3), mms_list):
            grad_b[:, j, i] += k_dict[i_sc][:, i].data * b_dict[i_sc][:, j].data

    # Gradient
    if method.lower() == "grad":
        out_data = grad_b

    # Divergence
    elif method.lower() == "div":
        out_data = np.zeros(len(b_dict["1"]))

        for mms_id in mms_list:
            out_data += dot(k_dict[mms_id], b_dict[mms_id]).data

    # Curl
    elif method.lower() == "curl":
        out_data = np.zeros((len(b_dict["1"]), 3))

        for mms_id in mms_list:
            out_data += cross(k_dict[mms_id], b_dict[mms_id]).data

    # B.div(B)
    elif method.lower() == "bdivb":
        out_data = np.zeros(b_avg.shape)

        for i in range(3):
            out_data[:, i] = np.sum(b_avg.data * grad_b[:, i, :], axis=1)

    # Curvature
    else:
        b_hat_list = [normalize(b) for b in b_list]

        out_data = c_4_grad(r_list, b_hat_list, method="bdivb").data

    out = _to_ts(out_data, b_dict)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .ts_skymap import ts_skymap

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def average_vdf(vdf, n_pts):
    r"""Time averages the velocity distribution functions over `n_pts` in time.

    Parameters
    ----------
    vdf : xarray.DataArray
        Time series of the velocity distribution function.
    n_pts : int
        Number of points (samples) of the averaging window.

    Returns
    -------
    vdf_avg : xarray.DataArray
        Time series of the time averaged velocity distribution function.

    """

    # Check input type
    assert isinstance(vdf, xr.Dataset), "vdf must be a xarray.Dataset"

    assert n_pts % 2 != 0, "The number of distributions to be averaged must be an odd"

    n_vdf = len(vdf.time.data)
    times = vdf.time.data

    pad_value = np.floor(n_pts / 2)
    avg_inds = np.arange(pad_value, n_vdf - pad_value, n_pts, dtype=int)
    time_avg = times[avg_inds]

    energy_avg = np.zeros((len(avg_inds), vdf.data.shape[1]))
    phi_avg = np.zeros((len(avg_inds), vdf.data.shape[2]))
    vdf_avg = np.zeros((len(avg_inds), *vdf.data.shape[1:]))

    for i, avg_ind in enumerate(avg_inds):
        l_bound = int(avg_ind - pad_value)
        r_bound = int(avg_ind + pad_value)
        vdf_avg[i, ...] = np.nanmean(
            vdf.data.data[l_bound:r_bound, ...],
            axis=0,
        )
        energy_avg[i, ...] = np.nanmean(
            vdf.energy.data[l_bound:r_bound, ...],
            axis=0,
        )
        phi_avg[i, ...] = np.nanmean(
            vdf.phi.data[l_bound:r_bound, ...],
            axis=0,
        )

    # Attributes
    glob_attrs = vdf.attrs  # Global attributes
    vdf_attrs = vdf.data.attrs  # VDF attributes
    coords_attrs = {k: vdf[k].attrs for k in ["time", "energy", "phi", "theta"]}

    # Get delta energy in global attributes for selected timestamps
    if "delta_energy_minus" in glob_attrs:
        glob_attrs["delta_energy_minus"] = glob_attrs["delta_energy_minus"][avg_inds, :]

    if "delta_energy_plus" in glob_attrs:
        glob_attrs["delta_energy_plus"] = glob_attrs["delta_energy_plus"][avg_inds, :]

    glob_attrs["esteptable"] = glob_attrs["esteptable"][: len(avg_inds)]

    vdf_avg = ts_skymap(
        time_avg,
        vdf_avg,
        energy_avg,
        phi_avg,
        vdf.theta.data,
        energy0=glob_attrs["energy0"],
        energy1=glob_attrs["energy1"],
        esteptable=glob_attrs["esteptable"][: len(avg_inds)],
        attrs=vdf_attrs,
        coords_attrs=coords_attrs,
        glob_attrs=glob_attrs,
    )

    return vdf_avg



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def mean_field(inp, deg):
    r"""Estimates mean field xm and wave field xw using polynomial fit of order
    `deg` for the number of columns larger than 3 assume that first column
    is time.

    Parameters
    ----------
    inp : array_like
        Input data.
    deg : int
        Degree of the fitting polynomial

    Returns
    -------
    inp_mean : numpy.ndarray
        Mean field.
    inp_wave : numpy.ndarray
        Wave field

    """
    inp_mean = inp
    inp_wave = inp

    n_rows, n_cols = inp.shape

    if n_cols >= 4:
        time = (inp[:, 0] - inp[0, 0]) / (inp[n_rows, 0] - inp[0, 0])
        data = inp[:, 1:]

    else:
        time = np.arange(len(inp))
        data = inp

    for i in range(data.shape[1]):
        res = np.polyfit(time, data[:, i], deg)
        polynomial_coeffs = res[0]
        inp_mean[:, i] = np.polyval(polynomial_coeffs, time)
        inp_wave[:, i] = data[:, i] - inp_mean[:, i]

    return inp_mean, inp_wave



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Built-in imports
from typing import Union

# 3rd party imports
import numpy as np
from scipy import constants, special

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _spherical_body(z, u, x, i_p):
    j_thermal = np.zeros(u.shape)

    if z > 0:
        j_thermal[u >= 0] = i_p * np.exp(-x[u >= 0])
        j_thermal[u < 0] = i_p * (1 - x[u < 0])
    elif z < 0:
        j_thermal[u >= 0] = i_p * (1 + x[u >= 0])
        j_thermal[u < 0] = i_p * np.exp(x[u < 0])

    return j_thermal


def _cylindrical_body(z, u, x, i_p):
    sq = np.zeros(u.shape)
    j_thermal = np.zeros(u.shape)

    sq[u < 0] = np.sqrt(abs(-x[u < 0]))
    sq[u >= 0] = np.sqrt(abs(+x[u >= 0]))
    erfv = special.erf(sq)

    if z > 0:
        j_thermal[u >= 0] = i_p * np.exp(-x[u >= 0])
        c_0 = (2.0 / np.sqrt(np.pi)) * sq[u < 0]
        c_1 = np.exp(-x[u < 0]) * (1.0 - erfv[u < 0])
        j_thermal[u < 0] = i_p * (c_0 + c_1)
    elif z < 0:
        j_thermal[u < 0] = i_p * np.exp(x[u < 0])

        c_0 = (2.0 / np.sqrt(np.pi)) * sq[u >= 0]
        c_1 = np.exp(x[u >= 0]) * (1.0 - erfv[u >= 0])
        j_thermal[u >= 0] = i_p * (c_0 + c_1)

    return j_thermal


def thermal_current(
    n: float,
    t: float,
    m: float,
    v: float,
    z: float,
    u: Union[float, np.ndarray],
    a: float,
    p_type: str,
) -> Union[float, np.ndarray]:
    r"""Calculates the thermal probe current to/from a cylindrical or
    spherical body, e.g. a Langmuir probe or the a spherical (cylindrical) S/C.

    Parameters
    ----------
    n : float
        Number density [m^3].
    t : float
        Temperature [K]
    m : float
        Mass [kg].
    z : {-1, 1}
        Charge
    v : float
        Velocity of the body with respect to the plasma [m/s].
    u : float or numpy.ndarray
        Body potential [V]
    a : float
        Area of body [m^2].
    p_type : {"sphere", "cylinder"}
        Probe type.

    Returns
    -------
    j_thermal : float or numpy.ndarray

    """

    assert p_type.lower() in ["sphere", "cylinder"]

    u = np.atleast_1d(u)

    # If zero density return zero current
    if n == 0 or t == 0:
        return None

    # Is the body moving with a velocity, V, with respect to the plasma ?
    # Criteria set such that it is considered important if V > 0.1 * V_th.
    if v < 0.1 * np.sqrt(constants.Boltzmann * t / m):
        # Ratio of potential to thermal energy.
        x = constants.elementary_charge * u / (constants.Boltzmann * t)

        # Total current to/from body.
        i_p = np.sqrt(t * constants.Boltzmann / (2.0 * np.pi * m))
        i_p *= a * n * constants.elementary_charge

    else:
        x = (
            constants.elementary_charge / (m * v**2 / 2 + constants.Boltzmann * t)
        ) * u
        i_p = np.sqrt(
            v**2 / 16 + t * constants.Boltzmann / (2.0 * np.pi * m),
        )
        i_p *= a * n * constants.elementary_charge

    if p_type == "sphere":
        # Spherical body case.
        j_thermal = _spherical_body(z, u, x, i_p)

    else:
        # Cylindrical body case.
        j_thermal = _cylindrical_body(z, u, x, i_p)

    return j_thermal



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

# Local imports
from .resample import resample
from .ts_scalar import ts_scalar
from .ts_vec_xyz import ts_vec_xyz

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def _check_method(flag_method):
    default_value = 0
    if flag_method.lower() == "e_perp+nan":
        default_value = np.nan

        flag_method = "e.b=0"

    assert flag_method.lower() in ["e.b=0", "e_par"], "Invalid flag"

    return flag_method, default_value


def edb(e_xyz, b_bgd, angle_lim: float = 20.0, flag_method: str = "E.B=0"):
    r"""Compute Ez under assumption :math:`\mathbf{E}.\mathbf{B}=0` or
    :math:`\mathbf{E}.\mathbf{B} \approx 0`

    Parameters
    ----------
    e_xyz : xarray.DataArray
        Time series of the electric field.
    b_bgd : xarray.DataArray
        Time series of the background magnetic field.
    angle_lim : float, Optional
        B angle with respect to the spin plane should be less than angle_lim
        degrees otherwise Ez is set to 0. Default is 20.
    flag_method : str, Optional
        Assumption on the direction of the measured electric field :
            "e.b=0" :  :math:`\mathbf{E}.\mathbf{B}=0`. (Default)
            "e_par" :  :math:`\mathbf{E}` field along the B projection is
            coming from parallelelectric field.
            "e_perp+nan" : to fill.

    Returns
    -------
    ed : xarray.DataArray
        Time series of the electric field output.
    d : xarray.DataArray
        Time series of the B elevation angle above spin plane.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2019-09-14T07:54:00.000","2019-09-14T08:11:00.000"]

    Spacecraft indices

    >>> mms_id = 1

    Load magnetic field, electric field and spacecraft position

    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_fast_l2", tint, mms_id)

    Compute Ez

    >>> e_z, alpha = pyrf.edb(e_xyz, b_xyz)

    """

    assert isinstance(e_xyz, xr.DataArray), "e_xyz must be a xarray.DataArray"
    assert isinstance(b_bgd, xr.DataArray), "b_bgd must be a xarray.DataArray"
    assert isinstance(angle_lim, (int, float)), "angle_lime must be int or float"

    flag_method, default_value = _check_method(flag_method)

    # Make sure that background magnetic field sampling matches the
    # electric field sampling
    b_bgd = resample(b_bgd, e_xyz)

    b_data = b_bgd.data
    e_data = e_xyz.data
    e_data[:, -1] *= default_value

    if flag_method.lower() == "e.b=0":
        # Calculate using assumption E.B=0
        b_angle = np.arctan2(
            b_data[:, 2],
            np.linalg.norm(b_data[:, :2], axis=1),
        )
        b_angle = np.rad2deg(b_angle)
        ind = np.abs(b_angle) > angle_lim

        if True in ind:
            e_data[ind, 2] = -np.sum(
                e_data[ind, :2] * b_data[ind, :2],
                axis=1,
            )
            e_data[ind, 2] /= b_data[ind, 2]

    else:
        # Calculate using assumption that E field along the B projection is
        # coming from parallel electric field
        b_angle = np.arctan2(
            b_data[:, 2],
            np.linalg.norm(b_data[:, :2], axis=1),
        )
        b_angle = np.rad2deg(b_angle)
        ind = np.abs(b_angle) < angle_lim

        if True in ind:
            e_data[ind, 2] = np.sum(e_data[ind, :2] * b_data[ind, :2], axis=1)
            e_data[ind, 2] *= b_data[ind, 2]
            e_data[ind, 2] /= np.linalg.norm(b_data[ind, :2], axis=1) ** 2

    b_angle = ts_scalar(e_xyz.time.data, b_angle, {"UNITS": "degrees"})
    e_data = ts_vec_xyz(
        e_xyz.time.data,
        e_data,
        e_xyz.attrs,
    )

    return e_data, b_angle



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# 3rd party imports
import numpy as np
import xarray as xr

__author__ = "Louis Richard"
__email__ = "louisr@irfu.se"
__copyright__ = "Copyright 2020-2023"
__license__ = "MIT"
__version__ = "2.4.2"
__status__ = "Prototype"


def integrate(inp, time_step: float = None):
    r"""Integrate time series.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the variable to integrate.

    time_step : float, Optional
        Time steps threshold. All time_steps larger than 3*time_step
        are assumed data gaps, default is that time_step is the
        smallest value of all time_steps of the time series.

    Returns
    -------
    out : xarray.DataArray
        Time series of the time integrated input.

    Examples
    --------
    >>> from pyrfu import mms, pyrf

    Time interval

    >>> tint = ["2015-12-14T01:17:40.200", "2015-12-14T01:17:41.500"]

    Spacecraft index

    >>> mms_id = 1

    Load magnetic field and electric field

    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_brst_l2", tint, mms_id)

    Convert electric field to field aligned coordinates

    >>> e_xyzfac = pyrf.convert_fac(e_xyz, b_xyz, [1, 0, 0])

    """

    assert isinstance(inp, xr.DataArray), "inp must be xarray.DataArray"

    time_tmp = inp.time.data.astype(np.float64) * 1e-9

    if inp.data.ndim == 1:
        data_tmp = inp.data[:, np.newaxis]
    else:
        data_tmp = inp.data

    data = np.transpose(np.vstack([time_tmp, np.transpose(data_tmp)]))

    delta_t = np.hstack([0, np.diff(data[:, 0])])

    if time_step is None:
        time_steps = np.diff(data[:, 0])

        # remove the smallest time step in case some problems
        time_step = np.min(np.delete(time_steps, np.argmin(time_steps)))

    delta_t[delta_t > 3 * time_step] = 0

    x_int = data
    for j in range(1, x_int.shape[1]):
        j_ok = ~np.isnan(x_int[:, j])

        x_int[j_ok, j] = np.cumsum(data[j_ok, j] * delta_t[j_ok])

    out = xr.DataArray(np.squeeze(data[:, 1:]), coords=inp.coords, dims=inp.dims)

    return out
